"""
    This class is used to connect the PTRAIL GUI to PTRAIL
    backend. All the GUI's functionalities are handled in this
    class.

    | Authors: Yaksh J Haranwala, Salman Haidri
"""
from distutils import util
import random
import re

import folium
import inspect
import pandas as pd

# GUI Imports.
from PyQt5 import QtWidgets, QtWebEngineWidgets, QtGui, QtCore
from ptrail.GUI.Table import TableModel
from ptrail.GUI.InputDialog import InputDialog
from ptrail.core.TrajectoryDF import PTRAILDataFrame

# Backend.
import ptrail.utilities.constants as const
from ptrail.features.kinematic_features import KinematicFeatures
from ptrail.features.temporal_features import TemporalFeatures
from ptrail.preprocessing.statistics import Statistics
from ptrail.preprocessing.filters import Filters
from ptrail.preprocessing.interpolation import Interpolation

# Statistics imports.
import matplotlib.pyplot as plt
import seaborn as sns
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar
from sklearn.feature_selection import mutual_info_classif


class GuiHandler:
    def __init__(self, filename, window):
        self.generateStats = False
        self.generateFeatureImportanceBtn = False
        self.statCanvas = None
        self.statFigure = None

        self.featureCanvas = None
        self.featureFigure = None

        self.ax = None
        self.traj_id_list = None

        self.map = None
        self._window = window
        self._data = None
        self._map_data = None
        self._model = None
        self._table = None

        self.display_df(filename=filename)

    def display_df(self, filename):
        """
            Display the DataFrame on the DFPane of the GUI.

            Parameters
            ----------
                filename: str
                    The name of the file. This is obtained from the GUI.

            Raises
            ------
                AttributeError:
                    If the user gives incorrect column names, then we ask
                    the user to enter them again.
        """
        try:
            self._window.statusBar.showMessage("Loading the Dataset...")

            # First, we clear out the DF pane area.
            # This is done in order to make sure that 2 dataframes
            # are not loaded simultaneously making the view cluttered.
            for i in reversed(range(self._window.DFPane.count())):
                item = self._window.DFPane.itemAt(i)
                if isinstance(item, QtWidgets.QTableView):
                    item.widget().close()

                # remove the item from layout
                self._window.DFPane.removeItem(item)

            col_names = self._get_input_params(labels=['Trajectory ID: ', 'DateTime: ', 'Latitude: ', 'Longitude: '],
                                               title="Enter Column Names",
                                               placeHolder=['Name of Identifier column', 'Name of Timestamp column',
                                                            'Name of Latitude column', 'Name of Longitude column'])
            if col_names is not None and col_names[0] != '' and len(col_names) == 4:
                # Read the data into a PTRAIL dataframe.
                self._data = PTRAILDataFrame(data_set=pd.read_csv(filename),
                                             traj_id=col_names[0].strip(),
                                             datetime=col_names[1].strip(),
                                             latitude=col_names[2].strip(),
                                             longitude=col_names[3].strip())
                self._map_data = self._data

                # Set the table model and display the dataframe.
                self._table = QtWidgets.QTableView()

                # NOTE: whenever we update DFs, make sure to send the data after resetting
                #       index and setting inplace as False.
                self._model = TableModel(self._data.reset_index(inplace=False))
                self._table.setModel(self._model)
                self._window.add_df_controller()
                self._window.DFPane.addWidget(self._table)
                self._window.statusBar.showMessage("Dataset Loaded Successfully.")

                # Get all the unique trajectory ids.
                ids_ = list(self._data.reset_index()['traj_id'].value_counts().keys())

                # Initiate the map placeholder.
                self.map = QtWebEngineWidgets.QWebEngineView()

                # Create the drop-down list for ID selection.
                self.traj_id_list = QtWidgets.QComboBox()
                self.traj_id_list.setFont(QtGui.QFont('Tahoma', 12))
                self.traj_id_list.addItems(ids_)

                # Add the drop-down and the map pane to the area.
                self._window.MapPane.addWidget(self.traj_id_list)
                self._window.MapPane.addWidget(self.map)

                # Actually draw the map.
                to_plot = self._map_data.reset_index().loc[self._map_data.reset_index()['traj_id']
                                                           == self.traj_id_list.currentText()]
                self._window.open_btn.deleteLater()
                self._draw_map(to_plot)
                self.draw_stats()
                self.add_column_drop_widget()
                self._window.runStatsBtn.setEnabled(True)
            else:
                self._window.open_file()

        except AttributeError or ValueError or TypeError:
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Critical)
            msg.setWindowTitle("Incorrect Column names")
            msg.setText("Incorrect Column names provided.\n"
                        "Please Enter the names again.")
            msg.exec()
            self.__init__(filename, self._window)

    def _draw_map(self, to_plot):
        self.map.setHtml('')

        # This is the colorblind palette taken from seaborn.
        colors = ['#0173b2', '#de8f05', '#029e73', '#d55e00', '#cc78bc',
                  '#ca9161', '#fbafe4', '#949494', '#ece133', '#56b4e9']

        sw = to_plot[['lat', 'lon']].min().values.tolist()
        ne = to_plot[['lat', 'lon']].max().values.tolist()

        # Create a map with the initial point.
        map_ = folium.Map(location=(to_plot[const.LAT].iloc[0], to_plot[const.LONG].iloc[0]),
                          zoom_start=13, tiles='CartoDB positron')

        # Then, create (lat, lon) pairs for the data points.
        locations = []
        for j in range(len(to_plot)):
            locations.append((to_plot['lat'].iloc[j], to_plot['lon'].iloc[j]))

        # Create text frame.
        iframe = folium.IFrame(f'<font size="1px">Trajectory ID: {self.traj_id_list.currentText()} ' + '<br>' +
                               f'Latitude: {locations[0][0]}' + '<br>' +
                               f'Longitude: {locations[0][1]} </font>')

        # Create start and end markers for the trajectory.
        popup = folium.Popup(iframe, min_width=180, max_width=200, max_height=75)

        folium.Marker([to_plot['lat'].iloc[0], to_plot['lon'].iloc[0]],
                      color='green',
                      popup=popup,
                      marker_color='green',
                      icon=folium.Icon(icon_color='green', icon='play', prefix='fa')).add_to(map_)

        # Create text frame.
        iframe = folium.IFrame(f'<font size="1px">Trajectory ID: {self.traj_id_list.currentText()} ' + '<br>' +
                               f'Latitude: {locations[-1][0]}' + '<br>' +
                               f'Longitude: {locations[-1][1]} </font>')

        # Create start and end markers for the trajectory.
        popup = folium.Popup(iframe, min_width=180, max_width=200, max_height=75)

        folium.Marker([to_plot['lat'].iloc[-1], to_plot['lon'].iloc[-1]],
                      color='red',
                      popup=popup,
                      marker_color='red',
                      icon=folium.Icon(icon_color='red', icon='stop', prefix='fa')).add_to(map_)

        # Add trajectory to map.
        folium.PolyLine(locations,
                        color=colors[random.randint(0, len(colors) - 1)],
                        ).add_to(map_)

        map_.fit_bounds([sw, ne])
        self.traj_id_list.currentIndexChanged.connect(lambda: self.redraw_map())
        self.map.setHtml(map_.get_root().render())

    def redraw_map(self):
        """
            Redraw the map when the traj_id is changed from the DropDown list.
        """
        # Check whether the QComboBox is empty or not. If so, don't redraw.
        # NOTE: This is done specifically to handle the case of filtering where
        #       some trajectories might be filtered out, and we have to update the
        #       id selection list.
        if self.traj_id_list.currentText() and self.traj_id_list.currentText() != '' and len(self._data) > 0:
            to_plot = self._map_data.reset_index().loc[
                self._map_data.reset_index()['traj_id'] == self.traj_id_list.currentText()]
            self._draw_map(to_plot)
            if self.generateStats:
                self.redraw_stat()

    def draw_stats(self):
        """
            Handle the objects of the statistics pane from here.
        """
        # Create the Stat Selection Drop down button.
        self._window.selectStatDropdown = QtWidgets.QComboBox()
        self._window.selectStatDropdown.currentIndexChanged.connect(lambda: self.redraw_stat())
        self._window.selectStatDropdown.setFont(QtGui.QFont("Tahoma", 12))

        # Create the matplotlib figure and axis for stats.
        self.statFigure = plt.figure()
        self.statCanvas = FigureCanvas(self.statFigure)
        self.statCanvas.setSizePolicy(
            QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Ignored)
        )

        # Create the matplotlib figure and axis for feature importance.
        self.featureFigure = plt.figure()
        self.featureCanvas = FigureCanvas(self.featureFigure)
        self.featureCanvas.setSizePolicy(
            QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Ignored)
        )

        # Add the figure and canvas to a new layout and finally add the layout to the StatsPane.
        self.generateFeatureImportanceBtn = QtWidgets.QPushButton("Generate Feature Importance")
        self.generateFeatureImportanceBtn.clicked.connect(lambda: self.generate_feature_imp_plot())
        self.generateFeatureImportanceBtn.setFont(QtGui.QFont("Tahoma", 12))
        self.generateFeatureImportanceBtn.setEnabled(False)
        new_draw_layout = QtWidgets.QVBoxLayout()
        new_draw_layout.addWidget(self._window.selectStatDropdown)
        new_draw_layout.addWidget(self.statCanvas)
        new_draw_layout.addWidget(self.featureCanvas)
        new_draw_layout.addWidget(self.generateFeatureImportanceBtn)
        self._window.StatsPane.addLayout(new_draw_layout)

    def generate_feature_imp_plot(self):
        """
            Take the input from the user and draw the mutual info
            plot.
        """
        try:
            args = self._get_input_params(['Class-Label Column', 'Number of Features', 'Segment Based'],
                                          title="Enter Classification Target column name",
                                          placeHolder=['Classification target column name', 'Number Features Wanted',
                                                       'True/False'])
            if args:
                feat, importance = None, None
                # If the user wants mutual info for point-based data, use the if-block.
                # For segment-based data, the else block is used.
                if not bool(util.strtobool(args[2].strip())):
                    # Generate the features and drop duplicate cols.
                    df = KinematicFeatures.generate_kinematic_features(self._data).dropna()
                    df = df.loc[:, ~df.columns.duplicated()]
                    df = df[[
                        'lat', 'lon', 'Distance', 'Distance_from_start', 'Speed', 'Acceleration', 'Jerk',
                        'Bearing', 'Bearing_Rate', 'Rate_of_bearing_rate', args[0]
                    ]]
                    Y = df[args[0].strip()]
                    df = df.drop(columns=[args[0]])
                else:
                    df = Statistics.generate_kinematic_stats(self._data, args[0], False)
                    df = Statistics.pivot_stats_df(df, args[0], False)
                    df = df.loc[:, ~df.columns.duplicated()]
                    Y = df[[args[0].strip()]]
                    df = df.drop(columns=[args[0]])

                # Check whether the number of features that the user asked is lesser
                # than actual number of features.
                num_features = int(args[1]) if len(df.columns) > int(args[1]) else len(df.columns)

                # Predict the mutual info.
                importance = mutual_info_classif(df, Y)
                feat = pd.Series(importance, df.columns).sort_values(ascending=False)
                feat = feat[0: num_features]

                # Plot the bar plots to the tool.
                self.featureFigure.clear()
                ax = self.featureFigure.add_subplot(111)

                feat.plot(kind='barh', color='skyblue', ax=ax)
                ax.set_title("Classification Feature Importance")
                ax.set_xlabel("Mutual Info")

                self.featureFigure.tight_layout()
                self.featureCanvas.draw()
                self.generateFeatureImportanceBtn.setText("Refresh Importance")
        except:
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Critical)
            msg.setWindowTitle("Not a classification problem")
            msg.setText("Feature importance is only shown for datasets that have a "
                        "target column.")
            msg.exec()

    def redraw_stat(self):
        """
            Redraw the statistics plot when the user changes the option from
            the Dropdown menu.
        """
        selected_stat = self._window.selectStatDropdown.currentText()
        stat_data = self._data.describe(percentiles=[0.1, 0.25, 0.5, 0.75, 0.9])
        colors = sns.color_palette('tab10')

        # Get the necessary stats out.
        percent_10 = stat_data[selected_stat]['10%']
        percent_25 = stat_data[selected_stat]['25%']
        percent_50 = stat_data[selected_stat]['50%']
        percent_75 = stat_data[selected_stat]['75%']
        percent_90 = stat_data[selected_stat]['90%']
        avg = stat_data[selected_stat]['mean']

        # Clear the figure and add an axes to it.
        self.statFigure.clear()
        ax = self.statFigure.add_subplot(111)

        # Attributes for the plot.
        text = ['10%', '25%', 'Median', '75%', '90%', 'Mean']
        horizontal_lines = [percent_10, percent_25, percent_50, percent_75, percent_90, avg]

        # Plot the line-plot and the stat lines.
        one_animal = self._data.reset_index().loc[self._data.reset_index()['traj_id']
                                                  == self.traj_id_list.currentText()]
        sns.lineplot(data=one_animal.reset_index(), x='DateTime', y=selected_stat, ax=ax, color='skyblue')
        for i in range(len(horizontal_lines)):
            ax.axhline(horizontal_lines[i], c=colors[i], linestyle='--', label=text[i])

        # Add the legend, rotate the ticks, set tight layout and draw it on the canvas.
        ax.legend(loc='upper right')
        plt.setp(ax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor")
        self.statFigure.tight_layout()
        self.statCanvas.draw()

    def add_column_drop_widget(self):
        """
            Add a List Widget to drop columns from the dataset. This
            widget is added to the CommandPalette.

            Note
            ----
                It is to be noted that the following columns are mandatory for
                PTrailDataFrame:
                    | 1. traj_id
                    | 2. DateTime
                    | 3. lat
                    | 4. lon

                Hence, these columns are not presented as options for deletion.
        """
        # A layout for containing the drop column setup.
        small_layout = QtWidgets.QVBoxLayout()

        # Create the list widget and add the options.
        self._window.dropColumnWidget = QtWidgets.QListWidget()
        self._window.dropColumnWidget.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self._window.dropColumnWidget.setFont(QtGui.QFont("Tahoma", 12))
        self._window.dropColumnWidget.setSizePolicy(
            QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Ignored))
        to_add = list(self._data.columns)
        to_add.remove('lat')
        to_add.remove('lon')
        self._window.dropColumnWidget.addItems(to_add)
        small_layout.addWidget(self._window.dropColumnWidget)

        # Add the button to drop the column.
        self._window.dropColumnBtn = QtWidgets.QPushButton("Drop Column(s)")
        self._window.dropColumnBtn.setFont(QtGui.QFont("Tahoma", 12))
        self._window.dropColumnBtn.clicked.connect(lambda: self.drop_col())
        small_layout.addWidget(self._window.dropColumnBtn)

        # Add the VBoxLayout to the Command Palette.
        self._window.CommandPalette.addLayout(small_layout)

    def run_command(self):
        """
            When the user pushes the Run button, run the user's
            selected function on the dataset.

            Returns
            -------
                None
        """
        if self._window.featureType.currentIndex() == 0:
            self._window.statusBar.showMessage("Running Filters ...")
            self._run_filters()
        elif self._window.featureType.currentIndex() == 1:
            self.generateStats = False
            self._window.statusBar.showMessage("Running Interpolation ...")
            self._run_ip()
        elif self._window.featureType.currentIndex() == 2:
            self.generateStats = True
            self._window.statusBar.showMessage("Running Kinematic Features ...")
            self._run_kinematic()
        elif self._window.featureType.currentIndex() == 3:
            self._window.statusBar.showMessage("Running Statistics ...")
            self._run_stats()
        else:
            self._window.statusBar.showMessage("Running Temporal Features ...")
            self._run_temporal()
        self.update_dropCol_options()

    def _run_ip(self):
        """
            Helper function for running the Interpolation commands
            from the GUI.

            Returns
            -------
                None
        """
        selected_function = self._window.featureListWidget.selectedItems()[0].text()

        if selected_function == 'Linear Interpolation':
            params = inspect.getfullargspec(Interpolation.interpolate_position).args
            params.remove('dataframe')
            params.remove('ip_type')
            args = self._get_input_params(params, title="Enter Parameters",
                                          placeHolder=['Sampling rate (in seconds)',
                                                       'Name of the column that contains class label (Leave empty if '
                                                       'none)'])

            if args:
                self._data = Interpolation.interpolate_position(dataframe=self._data,
                                                                ip_type='linear',
                                                                sampling_rate=float(args[0].strip()),
                                                                class_label_col=args[1].strip())

        elif selected_function == 'Cubic Interpolation':
            params = inspect.getfullargspec(Interpolation.interpolate_position).args
            params.remove('dataframe')
            params.remove('ip_type')
            args = self._get_input_params(params, title="Enter Parameters",
                                          placeHolder=['Sampling rate (in seconds)',
                                                       'Name of the column that contains class label (Leave empty if '
                                                       'none)']
                                          )

            if args:
                self._data = Interpolation.interpolate_position(dataframe=self._data,
                                                                ip_type='cubic',
                                                                sampling_rate=float(args[0].strip()),
                                                                class_label_col=args[1].strip())

        elif selected_function == 'Kinematic Interpolation':
            params = inspect.getfullargspec(Interpolation.interpolate_position).args
            params.remove('dataframe')
            params.remove('ip_type')
            args = self._get_input_params(params, title="Enter Parameters",
                                          placeHolder=['Sampling rate (in seconds)',
                                                       'Name of the column that contains class label (Leave empty if '
                                                       'none)'])

            if args:
                self._data = Interpolation.interpolate_position(dataframe=self._data,
                                                                ip_type='kinematic',
                                                                sampling_rate=float(args[0].strip()),
                                                                class_label_col=args[1].strip())

        elif selected_function == 'Random-Walk Interpolation':
            params = inspect.getfullargspec(Interpolation.interpolate_position).args
            params.remove('dataframe')
            params.remove('ip_type')
            args = self._get_input_params(params, title="Enter Parameters",
                                          placeHolder=['Sampling rate (in seconds)',
                                                       'Name of the column that contains class label (Leave empty if '
                                                       'none)']
                                          )

            if args:
                self._data = Interpolation.interpolate_position(dataframe=self._data,
                                                                ip_type='random_walk',
                                                                sampling_rate=float(args[0].strip()),
                                                                class_label_col=args[1].strip())

        self._window.statusBar.showMessage("Task Done ...")

        # Finally, update the GUI with the updated DF received from the
        # function results. DO NOT FORGET THE reset_index(inplace=False).
        self._model = TableModel(self._data.reset_index(inplace=False))
        self._table.setModel(self._model)

        # Also, update the map.
        self._map_data = self._data
        self._window.selectStatDropdown.blockSignals(True)
        self._window.selectStatDropdown.clear()
        self.statFigure.clear()
        self.statCanvas.draw()
        self._window.selectStatDropdown.blockSignals(False)
        self.redraw_map()

    def _run_kinematic(self):
        """
            Helper function for running the Kinematic commands
            from the GUI.

            Returns
            -------
                None
        """
        selected_function = self._window.featureListWidget.selectedItems()[0].text()

        # Based on the function selected by the user and whether it contains any
        # user given parameters, we will go ahead and run those features and update
        # the GUI with the results.

        if selected_function == 'All Kinematic Features':
            self._data = KinematicFeatures.generate_kinematic_features(self._data)
            self._window.statusBar.showMessage("Done ...")

            # Update the select-stats dropdown.
            self._window.selectStatDropdown.blockSignals(True)
            self._window.selectStatDropdown.clear()
            self._window.selectStatDropdown.addItems([
                'Distance', 'Distance_from_start', 'Speed', 'Acceleration', 'Jerk',
                'Bearing', 'Bearing_Rate', 'Rate_of_bearing_rate',
            ])
            self.generateFeatureImportanceBtn.setEnabled(True)
            self._window.selectStatDropdown.blockSignals(False)

        elif selected_function == 'Distance':
            self._data = KinematicFeatures.create_distance_column(self._data)
            self._window.selectStatDropdown.addItems(['Distance'])

        elif selected_function == 'Distance from Start':
            self._data = KinematicFeatures.create_distance_from_start_column(self._data)
            self._window.selectStatDropdown.addItems(['Distance_from_start'])

        elif selected_function == 'Point within Range':
            params = inspect.getfullargspec(KinematicFeatures.create_point_within_range_column).args
            params.remove('dataframe')
            args = self._get_input_params(params, title="Enter Parameters",
                                          placeHolder=['Coordinates (lat1, lon1, lat2, lon2)',
                                                       'Max distance from coordinates (metres)'])

            # If the user provided the input params, then run the function, else
            # wait for the user to play their part.
            if args:
                # Use Regex to get all the digits from the coords input and convert
                # it to required tuple to feed into the method.
                temp = re.findall(r'\d+', args[0].strip())
                coords = tuple(map(int, temp))

                dist_range = float(args[1].strip())
                self._data = KinematicFeatures.create_point_within_range_column(self._data,
                                                                                coordinates=coords,
                                                                                dist_range=dist_range)

        elif selected_function == 'Distance from Co-ordinates':
            params = inspect.getfullargspec(KinematicFeatures.create_distance_from_point_column).args
            params.remove('dataframe')

            args = self._get_input_params(params, title="Enter Parameters",
                                          placeHolder=['Coordinates (lat1, lon1, lat2, lon2)'])
            # If the user provided the input params, then run the function, else
            # wait for the user to play their part.
            if args:
                # Use Regex to get all the digits from the coords input and convert
                # it to required tuple to feed into the method.
                temp = re.findall(r'\d+', args[0].strip())
                coords = tuple(map(int, temp))

                self._data = KinematicFeatures.create_distance_from_point_column(dataframe=self._data,
                                                                                 coordinates=coords)

        elif selected_function == 'Speed':
            self._data = KinematicFeatures.create_speed_column(self._data)
            self._window.selectStatDropdown.addItems(['Speed'])

        elif selected_function == 'Acceleration':
            self._data = KinematicFeatures.create_acceleration_column(self._data)
            self._window.selectStatDropdown.addItems(['Acceleration'])

        elif selected_function == 'Jerk':
            self._data = KinematicFeatures.create_jerk_column(self._data)
            self._window.selectStatDropdown.addItems(['Jerk'])

        elif selected_function == 'Bearing':
            self._data = KinematicFeatures.create_bearing_column(self._data)
            self._window.selectStatDropdown.addItems(['Bearing'])

        elif selected_function == 'Bearing Rate':
            self._data = KinematicFeatures.create_bearing_rate_column(self._data)
            self._window.selectStatDropdown.addItems(['Bearing_Rate'])

        elif selected_function == 'Rate of Bearing Rate':
            self._data = KinematicFeatures.create_rate_of_br_column(self._data)
            self._window.selectStatDropdown.addItems(['Rate_of_bearing_rate'])

        # Redraw the stats panel.
        self.redraw_stat()

        # Finally, update the GUI with the updated DF received from the
        # function results. DO NOT FORGET THE reset_index(inplace=False).
        self._map_data = self._data
        self._window.statusBar.showMessage("Task Done ...")
        self._model = TableModel(self._data.reset_index(inplace=False))
        self._table.setModel(self._model)

    def _run_temporal(self):
        """
            Helper function for running the Temporal commands
            from the GUI.

            Returns
            -------
                None
        """
        selected_function = self._window.featureListWidget.selectedItems()[0].text()

        if selected_function == "All Temporal Features":
            self._data = TemporalFeatures.generate_temporal_features(self._data)

        elif selected_function == "Date":
            self._data = TemporalFeatures.create_date_column(self._data)

        elif selected_function == "Time":
            self._data = TemporalFeatures.create_time_column(self._data)

        elif selected_function == "Day of the Week":
            self._data = TemporalFeatures.create_day_of_week_column(self._data)

        elif selected_function == "Weekend Indicator":
            self._data = TemporalFeatures.create_weekend_indicator_column(self._data)

        elif selected_function == "Time of Day":
            self._data = TemporalFeatures.create_time_of_day_column(self._data)

        # Finally, update the GUI with the updated DF received from the
        # function results. DO NOT FORGET THE reset_index(inplace=False).
        self._map_data = self._data
        self._window.statusBar.showMessage("Task Done ...")
        self._model = TableModel(self._data.reset_index(inplace=False))
        self._table.setModel(self._model)

    def _run_filters(self):
        """
            Helper function for running the Filter commands
            from the GUI.

            Returns
            -------
                None
        """
        selected_function = self._window.featureListWidget.selectedItems()[0].text()

        if selected_function == 'Hampel Filter':
            params = inspect.getfullargspec(Filters.hampel_outlier_detection).args
            params.remove('dataframe')

            args = self._get_input_params(params, title="Enter Parameters",
                                          placeHolder=['Filter by Metric (Enter Column Name)'])
            # If the user provided the input params, then run the function, else
            # wait for the user to play their part.
            if args:
                self._data = Filters.hampel_outlier_detection(dataframe=self._data,
                                                              column_name=args[0].strip())

        elif selected_function == 'Remove Duplicates':
            self._data = Filters.remove_duplicates(self._data)

        elif selected_function == 'By Trajectory ID':
            params = inspect.getfullargspec(Filters.filter_by_traj_id).args
            params.remove('dataframe')

            args = self._get_input_params(params, title="Enter Parameters",
                                          placeHolder=['Trajectory ID'])
            # If the user provided the input params, then run the function, else
            # wait for the user to play their part.
            if args:
                self._data = Filters.filter_by_traj_id(dataframe=self._data,
                                                       traj_id=args[0].strip())
        elif selected_function == 'By Bounding Box':
            params = inspect.getfullargspec(Filters.filter_by_bounding_box).args
            params.remove('dataframe')

            args = self._get_input_params(params, title="Enter Parameters",
                                          placeHolder=["lat1, lon1, lat2, lon2",
                                                       "Points inside the bounding box? (True/False)"])
            # If the user provided the input params, then run the function, else
            # wait for the user to play their part.
            if args:
                temp = re.findall("\d+\.\d+", args[0])
                coords = tuple(map(float, temp))
                self._data = Filters.filter_by_bounding_box(dataframe=self._data,
                                                            bounding_box=coords,
                                                            inside=(bool(util.strtobool(args[1].strip()))))

        elif selected_function == 'By Date':
            params = inspect.getfullargspec(Filters.filter_by_date).args
            params.remove('dataframe')

            args = self._get_input_params(params, title="Enter Parameters",
                                          placeHolder=["YYYY-MM-DD", "YYYY-MM-DD"])
            # If the user provided the input params, then run the function, else
            # wait for the user to play their part.
            if args:
                if args[0] == '':
                    self._data = Filters.filter_by_date(dataframe=self._data,
                                                        end_date=args[1].strip())
                elif args[1] == '':
                    self._data = Filters.filter_by_date(dataframe=self._data,
                                                        start_date=args[0].strip())
                else:
                    self._data = Filters.filter_by_date(dataframe=self._data,
                                                        start_date=args[0].strip(),
                                                        end_date=args[1].strip())

        elif selected_function == 'By DateTime':
            params = inspect.getfullargspec(Filters.filter_by_datetime).args
            params.remove('dataframe')

            args = self._get_input_params(params, title="Enter Parameters",
                                          placeHolder=['YYYY-MM-DD hh:mm:ss', 'YYYY-MM-DD hh:mm:ss'])
            # If the user provided the input params, then run the function, else
            # wait for the user to play their part.
            if args:
                if args[0] == '':
                    self._data = Filters.filter_by_datetime(dataframe=self._data,
                                                            end_dateTime=args[1].strip())
                elif args[1] == '':
                    self._data = Filters.filter_by_datetime(dataframe=self._data,
                                                            start_dateTime=args[0].strip())
                else:
                    self._data = Filters.filter_by_datetime(dataframe=self._data,
                                                            start_dateTime=args[0].strip(),
                                                            end_dateTime=args[1].strip())

        elif selected_function == 'By Maximum Speed':
            params = inspect.getfullargspec(Filters.filter_by_max_speed).args
            params.remove('dataframe')

            args = self._get_input_params(params, title="Enter Parameters",
                                          placeHolder=['Maximum speed between consecutive points (metres/second)'])
            # If the user provided the input params, then run the function, else
            # wait for the user to play their part.
            if args:
                self._data = Filters.filter_by_max_speed(dataframe=self._data,
                                                         max_speed=float(args[0].strip()))

        elif selected_function == 'By Minimum Speed':
            params = inspect.getfullargspec(Filters.filter_by_min_speed).args
            params.remove('dataframe')

            args = self._get_input_params(params, title="Enter Parameters",
                                          placeHolder=['Minimum speed between consecutive points (metres/second)'])
            # If the user provided the input params, then run the function, else
            # wait for the user to play their part.
            if args:
                self._data = Filters.filter_by_min_speed(dataframe=self._data,
                                                         min_speed=float(args[0].strip()))

        elif selected_function == 'By Minimum Consecutive Distance':
            params = inspect.getfullargspec(Filters.filter_by_min_consecutive_distance).args
            params.remove('dataframe')

            args = self._get_input_params(params, title="Enter Parameters",
                                          placeHolder=['Minimum Distance between 2 points (metres)'])
            # If the user provided the input params, then run the function, else
            # wait for the user to play their part.
            if args:
                self._data = Filters.filter_by_min_consecutive_distance(dataframe=self._data,
                                                                        min_distance=float(args[0].strip()))

        elif selected_function == 'By Maximum Consecutive Distance':
            params = inspect.getfullargspec(Filters.filter_by_max_consecutive_distance).args
            params.remove('dataframe')

            args = self._get_input_params(params, title="Enter Parameters",
                                          placeHolder=['Max Distance between 2 points (metres)'])
            # If the user provided the input params, then run the function, else
            # wait for the user to play their part.
            if args:
                self._data = Filters.filter_by_max_consecutive_distance(dataframe=self._data,
                                                                        max_distance=float(args[0].strip()))

        elif selected_function == 'By Maximum Distance and Speed':
            params = inspect.getfullargspec(Filters.filter_by_max_distance_and_speed).args
            params.remove('dataframe')

            args = self._get_input_params(params, title="Enter Parameters",
                                          placeHolder=['Max Distance (metres)', 'Max Speed (metres/second)'])
            # If the user provided the input params, then run the function, else
            # wait for the user to play their part.
            if args:
                self._data = Filters.filter_by_max_distance_and_speed(dataframe=self._data,
                                                                      max_distance=float(args[0].strip()),
                                                                      max_speed=float(args[1].strip()))

        elif selected_function == 'By Minimum Distance and Speed':
            params = inspect.getfullargspec(Filters.filter_by_min_distance_and_speed).args
            params.remove('dataframe')

            args = self._get_input_params(params, title="Enter Parameters",
                                          placeHolder=['Minimum Distance (metres)', 'Minimum Speed (metres/second)'])
            # If the user provided the input params, then run the function, else
            # wait for the user to play their part.
            if args:
                self._data = Filters.filter_by_min_distance_and_speed(dataframe=self._data,
                                                                      min_distance=float(args[0].strip()),
                                                                      min_speed=float(args[1].strip()))

        elif selected_function == 'Remove Outliers by Consecutive Distance':
            self._data = Filters.filter_outliers_by_consecutive_distance(dataframe=self._data)

        elif selected_function == 'Remove Outliers by Consecutive Speed':
            self._data = Filters.filter_outliers_by_consecutive_speed(dataframe=self._data)

        elif selected_function == 'Remove Trajectories with Less Points':
            params = inspect.getfullargspec(Filters.remove_trajectories_with_less_points).args
            params.remove('dataframe')

            args = self._get_input_params(params, title="Enter Parameters",
                                          placeHolder=['Number of Minimum points in the trajectory to avoid removal'])
            # If the user provided the input params, then run the function, else
            # wait for the user to play their part.
            if args:
                self._data = Filters.remove_trajectories_with_less_points(dataframe=self._data,
                                                                          num_min_points=int(args[0].strip()))

        # Update the traj_id list in case if some trajectories have been completely
        # removed.
        self._map_data = self._data
        self.traj_id_list.clear()
        self.traj_id_list.addItems(list(self._data.reset_index()['traj_id'].value_counts().keys()))

        # Finally, update the GUI with the updated DF received from the
        # function results. DO NOT FORGET THE reset_index(inplace=False).
        self._window.statusBar.showMessage("Task Done ...")
        self._model = TableModel(self._data.reset_index(inplace=False))
        self._table.setModel(self._model)

    def _run_stats(self):
        """
            Helper function for running the Statistical commands
            from the GUI.

            Returns
            -------
                None
        """
        selected_function = self._window.featureListWidget.selectedItems()[0].text()

        if selected_function == 'Segment Trajectories':
            params = inspect.getfullargspec(Statistics.segment_traj_by_days).args
            params.remove('dataframe')

            args = self._get_input_params(params, title="Enter Parameters",
                                          placeHolder=['Duration of each segment in days'])
            # If the user provided the input params, then run the function, else
            # wait for the user to play their part.
            if args:
                self._data = Statistics.segment_traj_by_days(dataframe=self._data,
                                                             num_days=int(args[0].strip()))
                self._map_data = self._data
        elif selected_function == 'Generate Kinematic Statistics':
            params = inspect.getfullargspec(Statistics.generate_kinematic_stats).args
            params.remove('dataframe')

            args = self._get_input_params(params, title="Enter Parameters",
                                          placeHolder=['Class-label column name', 'Is the trajectory segmented?'])
            # If the user provided the input params, then run the function, else
            # wait for the user to play their part.
            if args:
                self._data = Statistics.generate_kinematic_stats(dataframe=self._data,
                                                                 target_col_name=args[0].strip(),
                                                                 segmented=bool(util.strtobool(args[1].strip())))

        elif selected_function == 'Pivot Statistics DF':
            params = inspect.getfullargspec(Statistics.pivot_stats_df).args
            params.remove('dataframe')

            args = self._get_input_params(params, title="Enter Parameters",
                                          placeHolder=['Class-label column name', 'Is the trajectory segmented?'])
            # If the user provided the input params, then run the function, else
            # wait for the user to play their part.
            if args:
                self._data = Statistics.pivot_stats_df(dataframe=self._data,
                                                       target_col_name=args[0].strip(),
                                                       segmented=bool(util.strtobool(args[1].strip())))

        # Finally, update the GUI with the updated DF received from the
        # function results. DO NOT FORGET THE reset_index(inplace=False).
        self._window.statusBar.showMessage("Task Done ...")
        self._model = TableModel(self._data.reset_index(inplace=False))
        self._table.setModel(self._model)

    def _get_input_params(self, labels, title, placeHolder):
        """
            Take the input parameters for the function in question from
            the user.

            Parameters
            ----------
                labels: list
                    The name of the parameters.
                title: str
                    The title of the input dialog box.
                placeHolder: list
                    Default text for each QLineEdit.

            Returns
            -------
                list:
                    A list containing the user input.
        """
        input_dialog = InputDialog(parent=self._window,
                                   labels=labels,
                                   title=title,
                                   placeHolders=placeHolder)
        if input_dialog.exec_():
            args = input_dialog.getInputs()

            return args

    def update_dropCol_options(self):
        """
            Update the options in the QListWidget for dropping the columns.
        """
        try:
            self._window.dropColumnWidget.clear()
            toAdd = list(self._data.columns)
            toAdd.remove('lat')
            toAdd.remove('lon')
            self._window.dropColumnWidget.addItems(toAdd)
        except ValueError:
            self._window.dropColumnWidget.clear()
            toAdd = list(self._data.columns)
            self._window.dropColumnWidget.addItems(toAdd)

    def drop_col(self):
        """
            Drop the columns based on the user selection.
        """
        # Get the selected items and then create a list
        # of the column names to drop.
        items = self._window.dropColumnWidget.selectedItems()

        to_drop = list()
        for it in items:
            to_drop.append(it.text())

        # Now, whenever we drop a column related to one of the stats,
        # then we go ahead and remove that option from the stats selection
        # panel as well.
        all_stat_items = [self._window.selectStatDropdown.itemText(i)
                          for i in range(self._window.selectStatDropdown.count())]
        for val in to_drop:
            if val in all_stat_items:
                all_stat_items.remove(val)

        # Make sure to block this signal before clearing and unblock
        # after adding new options :).
        self._window.selectStatDropdown.blockSignals(True)
        self._window.selectStatDropdown.clear()
        self._window.selectStatDropdown.addItems(all_stat_items)
        self._window.selectStatDropdown.blockSignals(False)

        # Drop the column(s) selected by the user.
        self._data.drop(columns=to_drop, inplace=True)

        # Update the options in the drop-column selection list and display the table again.
        self.update_dropCol_options()
        self._model = TableModel(self._data.reset_index(inplace=False))
        self._table.setModel(self._model)



"""
    This python module is the abstract definition of the Table view
    for viewing the dataframe inside the GUI.

    | Authors: Yaksh J Haranwala
"""

from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import Qt


class TableModel(QtCore.QAbstractTableModel):
    def __init__(self, data):
        super(TableModel, self).__init__()
        self._data = data

    def headerData(self, section, orientation, role=QtCore.Qt.DisplayRole):
        if role != QtCore.Qt.DisplayRole:
            return QtCore.QVariant()

        if orientation == QtCore.Qt.Horizontal:
            try:
                return self._data.columns.tolist()[section]
            except (IndexError,):
                return QtCore.QVariant()
        elif orientation == QtCore.Qt.Vertical:
            try:
                return self._data.index.tolist()[section]
            except (IndexError,):
                return QtCore.QVariant()

    #
    def data(self, index, role=Qt.DisplayRole):
        if index.isValid():
            if role == Qt.DisplayRole:
                return str(self._data.iloc[index.row(), index.column()])
        return None

    def rowCount(self, parent=None):
        return self._data.shape[0]

    def columnCount(self, parent=None):
        return self._data.shape[1]


"""
    The conversions modules contains various available methods
    that can be used to convert given data into another format.

    | Authors: Yaksh J Haranwala, Salman Haidri
"""
from typing import Text


class Conversions:
    @staticmethod
    def convert_directions_to_degree_lat_lon(data, latitude: Text, longitude: Text):
        """
            Convert the latitude and longitude format from degrees (NSEW)
            to float values. This is used for datasets like the Atlantic Hurricane dataset
            where the coordinates are not given as float values but are instead given as
            degrees.

            References
            ----------
                "Arina De Jesus Amador Monteiro Sanches. “Uma Arquitetura E Imple-menta ̧c ̃ao Do M ́odulo De
                Pr ́e-processamento Para Biblioteca Pymove”.Bachelor’s thesis. Universidade Federal Do Cear ́a, 2019"
        """

        def decimal_degree_to_decimal(col):
            if col[latitude][-1:] == 'N':
                col[latitude] = float(col[latitude][:-1])
            else:
                col[latitude] = float(col[latitude][:-1]) * -1

            if col[longitude][-1:] == 'E':
                col[longitude] = float(col[longitude][:-1])
            else:
                col[longitude] = float(col[longitude][:-1]) * -1 + 360 if float(col[longitude][:-1]) * -1 < -180 \
                    else float(col[longitude][:-1]) * -1
            return col

        return data.apply(decimal_degree_to_decimal, axis=1)



"""
    This File contains the visualization that is a Donut chart depicting the breakdown
    of animals by each pasture. The user can change the pasture to see the breakdown of
    individual pastures.

    Warning
    -------
        The visualizations in this module are currently developed with a focus around the
        starkey.csv data as it has been developed as a side project by the developers. It
        will further be integrated into the library as a general class of visualizers in
        the time to come. Some of the visualization types may or may not work with other
        datasets.

    | Authors: Yaksh J Haranwala
"""
import pandas as pd
import geopandas as gpd
from IPython.core.display import display
from ipywidgets import widgets
from shapely.geometry import Polygon
import matplotlib.pyplot as plt

from ptrail.core.TrajectoryDF import PTRAILDataFrame
from ptrail.preprocessing.filters import Filters as filt


class InteractiveDonut:
    # Class variables to store datasets and the widget.
    __habitat_data = None
    __traj_data = None
    __dropdown = None

    @staticmethod
    def animals_by_pasture(trajectories: PTRAILDataFrame, habitat: pd.DataFrame):
        """
            Plot a donut chart that shows the proportion of animals for each pasture.

            Parameters
            ----------
                trajectories: PTRAILDataFrame
                    The dataframe that contains trajectory data.
                habitat: pd.DataFrame
                    The dataframe that contains habitat data.

            Returns
            -------
                None
        """
        # Store the datasets in the class variables.
        InteractiveDonut.__traj_data = trajectories
        InteractiveDonut.__habitat_data = gpd.GeoDataFrame(habitat.reset_index(),
                                                           geometry=gpd.points_from_xy(habitat['lon'], habitat['lat']))

        # The list of available pastures.
        habitats = ['MDWCRK', 'SMITH-BALLY', 'STRIP', 'HORSE', 'BEAR', 'HALFMOON']

        # Create the dropdown widget.
        InteractiveDonut.__dropdown = widgets.Dropdown(options=habitats,
                                                       value='SMITH-BALLY',
                                                       description='Pasture',
                                                       disabled=False)

        # Show the plot.
        ie = widgets.interactive_output(InteractiveDonut.__plot_pasture_donut,
                                        {'pasture_name': InteractiveDonut.__dropdown})

        # Display the widget and its output side-by-side.
        display(ie, InteractiveDonut.__dropdown)

    @staticmethod
    def __plot_pasture_donut(pasture_name):
        """
            Plot the donut chart as per the pasture.

            Parameters
            ----------
                pasture_name: str
                    The name of the pasture for which the Donut chart is to be plotted.

            Returns
            -------
                None
        """
        small = InteractiveDonut.__habitat_data.loc[InteractiveDonut.__habitat_data['CowPast'] == pasture_name]
        animals = InteractiveDonut._get_count_by_pasture(small,
                                                         InteractiveDonut.__traj_data)

        deer, cattle, elk = 0, 0, 0
        for i in range(len(animals)):
            if 'D' in animals[i]:
                deer += 1
            elif 'E' in animals[i]:
                elk += 1
            else:
                cattle += 1

        # Convert the data created above to a dict.
        data = {'Deer': deer, 'Elk': elk, 'Cattle': cattle}

        # Sort the above dictionary by values.
        data = {k: v for k, v in sorted(data.items(), key=lambda item: item[1])}

        # Create a circle at the center of the plot
        my_circle = plt.Circle((0, 0), 0.65, color='white')

        # Add text to the centre of the plot.
        my_text = plt.Text(-0.3, -0.15, f'Breakdown of\n'
                                        f'  animals in\n' +
                           pasture_name)

        # Custom wedges
        plt.pie(x=data.values(), labels=data.keys(),
                wedgeprops={'linewidth': 2, 'edgecolor': 'white'},
                startangle=90)
        p = plt.gcf()
        p.gca().add_artist(my_circle)
        p.gca().add_artist(my_text)

        stats = InteractiveDonut.__get_pasture_stats(pasture_name)

        plt.gcf().text(0.9, 0.5, stats, fontsize=12)
        plt.tight_layout()

    @staticmethod
    def __get_pasture_stats(pasture: str):
        """
            Given the habitat dataset and the name of the pasture,
            return a text string containing the following information:
                | 1. Average Canopy Cover
                | 2. Average ground elevation
                | 3. Number of water bodies inside the pasture
                | 4. Average distance to nearest open road

            Parameters
            ----------
                pasture: str
                    The name of the pasture for which the stats are to be calculated.

            Returns
            -------
                str:
                    String containing the habitat stats.
        """
        # First, filter the dataset by pasture.
        pasture_df = InteractiveDonut.__habitat_data.loc[InteractiveDonut.__habitat_data['CowPast'] == pasture]

        # Calculate the average canopy cover of the pasture.
        mean_canopy = pasture_df['Canopy'].mean()

        # Calculate the average elevation.
        mean_elev = pasture_df['Elev'].mean()

        # Calculate the average distance to nearest open road.
        mean_open_dist = pasture_df['DistOPEN'].mean()

        # Calculate the number of water bodies inside the pasture.
        num_water = len(pasture_df.loc[pasture_df['DistEWat'] == 0])

        return f"{pasture} Description\n\n" \
               f"Average Canopy Cover: {round(mean_canopy, 2)} %\n" \
               f"Average Ground Elevation: {round(mean_elev, 2)} m\n" \
               f"Number of water bodies: {num_water} \n" \
               f"Average Distance to Nearest Road: {round(mean_open_dist, 2)} m\n"

    @staticmethod
    def _get_count_by_pasture(habitat: gpd.GeoDataFrame, trajectories: PTRAILDataFrame):
        """
            Filter the dataset by pasture and return the counts of deer, elk
            and cattle individually.

            Parameters
            ----------
                habitat: gpd.GeoDataFrame
                    The dataframe containing the habitat data.
                trajectories: PTRAILDataFrame
                    The dataframe containing the Trajectory data.

            Returns
            -------
                dict:
                    animal, count pairs.
        """
        # Using GeoPandas, get the bounding box of the pasture.
        bbox = habitat.geometry.total_bounds

        # Since GeoPandas uses lon, lat format, we need to swap it
        # in order to make it lat, lon format.
        bbox[0], bbox[1] = bbox[1], bbox[0]
        bbox[2], bbox[3] = bbox[3], bbox[2]
        bbox = tuple(bbox)

        # Using PTRAIL, filter the points that are inside the bounding box.
        filtered_df = filt.filter_by_bounding_box(dataframe=trajectories, bounding_box=bbox, inside=True)
        filtered_df = filtered_df.reset_index()

        return filtered_df.reset_index()['traj_id'].unique().tolist()

    @staticmethod
    def plot_area_donut(habitat: pd.DataFrame):
        """
            Given the trajectories and the habitat dataset, plot a donut plot
            which shows the area of each individual pasture as a ring and then
            has an interactive element that shows the distribution of animals
            upon clicking the pasture ring.

            Parameters
            ----------
                habitat: pd.core.dataframe.DataFrame
                    The dataset containing the habitat data.

            Returns
            -------
                None
        """
        # Get the area data for the habitat.
        areas = InteractiveDonut._get_pasture_area(habitat)
        areas = areas.sort_values(by='area')

        # Create a circle at the center of the plot
        my_circle = plt.Circle((0, 0), 0.65, color='white')

        # Add text to the centre of the plot.
        my_text = plt.Text(-0.4, -0.15, f'    Breakdown of\n'
                                        'Starkey Forest Area\n'
                                        '     by Pastures')

        # Custom wedges
        palette = ['#377eb8', '#ff7f00', '#4daf4a',
                   '#f781bf', '#a65628', '#984ea3',
                   '#999999', '#e41a1c', '#dede00']
        plt.pie(x=areas['area'], labels=areas['pasture'],
                wedgeprops={'linewidth': 2, 'edgecolor': 'white'},
                startangle=90, colors=palette, rotatelabels=True)
        p = plt.gcf()
        p.set_size_inches(7, 5)
        p.gca().add_artist(my_circle)
        p.gca().add_artist(my_text)

        plt.tight_layout()

    @staticmethod
    def _get_pasture_area(dataset: pd.DataFrame):
        """
            Given the dataset containing the habitat data, return a dataframe
            containing the name of the pasture and the area of the pasture.

            Note
            ----
                It was noted that the starkey dataset had a pasture that did not have
                any name, hence it was renamed to STARK and if a pasture has less than
                3 points of data, then it is dropped since it does not have enough data
                to calculate its area. Moreover, if a pasture has an area of 0, it is
                dropped furthermore to clean the data.

            Note
            ----
                The area calculated is in km^2.

            Parameters
            ----------
                dataset: PTRAILDataFrame
                    The dataframe containing habitat data.

            Returns
            -------
                pd.core.dataframe.DataFrame:
                    The pandas dataframe containing the name of the pastures and their
                    respective areas.
        """
        # Convert the dataframe given into a GeoDataFrame.
        habitat_gdf = gpd.GeoDataFrame(dataset.reset_index(),
                                       geometry=gpd.points_from_xy(dataset['lon'], dataset['lat']))

        # Set the crs to EPSG:4326 and then make sure that it projected to EPSG:3857
        # since we want the area to be in metres.
        habitat_gdf.crs = "EPSG:4326"
        habitat_gdf = habitat_gdf.to_crs('EPSG:3857')

        # Get a list of all the unique habitats.
        habitats = habitat_gdf['CowPast'].unique()

        df = pd.DataFrame(columns=['pasture', 'area'])
        for val in habitats:
            # the try catch does the job of ignoring the pasture with less
            # than 2 points.
            try:
                small = habitat_gdf.loc[habitat_gdf['CowPast'] == val]

                # Rename the nan pasture to STARK
                if type(val) != str:
                    continue

                # Calculate the area and then append it to the dataframe.
                df.loc[val] = Polygon(small['geometry'].tolist()).area / 10e6
            except ValueError:
                continue

        # Clear the trash out of the DF and return it.
        df = df.reset_index().drop(columns=['pasture'])
        df = df.rename(columns={'index': 'pasture'})

        return df



"""
    This file launches PTRAIL's GUI module.

    | Author: Yaksh J Haranwala
"""
from PyQt5 import QtWidgets
import sys
from ptrail.GUI.gui import Ui_MainWindow

if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    MainWindow = QtWidgets.QMainWindow()
    ui = Ui_MainWindow(MainWindow)
    MainWindow.showMaximized()
    sys.exit(app.exec_())



"""
    The Datasets.py module is used to load built-in datasets to variables.
    All the datasets loaded are stored and returned in a PTRAILDataFrame
    Currently, the library has the following datasets available to use:

        | 1. Atlantic Hurricanes Dataset
        | 2. Traffic Dataset (a smaller subset)
        | 3. Geo-life Dataset (a smaller subset)
        | 4. Seagulls Dataset
        | 5. Ships Dataset (a smaller subset)
        | 6. Starkey Animals Dataset
        | 7. Starkey Habitat Dataset (accompanies the starkey dataset)

        Note
        ----
            The Starkey Habitat Dataset is not loaded is not loaded into a PTrailDataframe since
            it is not a movement dataset and rather contains contextual information about the starkey
            habitat. It is rather loaded into a pandas dataframe and returned as is.

    | Authors: Yaksh J Haranwala
"""
import pandas as pd
from ptrail.core.TrajectoryDF import PTRAILDataFrame


class Datasets:
    @staticmethod
    def load_hurricanes():
        """
            Load the Atlantic Hurricane dataset into the PTRAILDataFrame
            and return it.

            Returns
            -------
                PTRAILDataFrame:
                    The atlantic hurricanes dataset loaded into a PTrailDataFrame.
        """
        # read the CSV file from the repository using pandas.
        df = pd.read_csv('https://raw.githubusercontent.com/YakshHaranwala/PTRAIL/main/examples/data'
                         '/atlantic_hurricanes.csv')

        # Load the dataset into a PTrailDataFrame and print the dataframe
        # information.
        to_return = PTRAILDataFrame(data_set=df,
                                    latitude='lat',
                                    longitude='long',
                                    datetime='DateTime',
                                    traj_id='traj_id',
                                    rest_of_columns=[])
        print(to_return)

        # return the PTRAILDataFrame
        return to_return

    @staticmethod
    def load_traffic_data():
        """
            Load the Traffic dataset into the PTRAILDataFrame and return it.

            Returns
            -------
                PTRAILDataFrame:
                    The traffic dataset loaded into a PTrailDataFrame.
        """
        # read the CSV file from the repository using pandas.
        df = pd.read_csv('https://raw.githubusercontent.com/YakshHaranwala/PTRAIL/main/examples/data/car_traffic.csv')

        # Load the dataset into a PTrailDataFrame and print the dataframe
        # information.
        to_return = PTRAILDataFrame(data_set=df,
                                    latitude='lat',
                                    longitude='long',
                                    datetime='DateTime',
                                    traj_id='traj_id',
                                    rest_of_columns=[])
        print(to_return)

        # return the PTRAILDataFrame
        return to_return

    @staticmethod
    def load_geo_life_sample():
        """
            Load the Geo-Life Sample dataset into the PTRAILDataFrame and return it.

            Returns
            -------
                PTRAILDataFrame:
                    The geo-life sample dataset loaded into a PTrailDataFrame.
        """
        # read the CSV file from the repository using pandas.
        df = pd.read_csv(
            'https://raw.githubusercontent.com/YakshHaranwala/PTRAIL/main/examples/data/geolife_sample.csv'
        )

        # Load the dataset into a PTrailDataFrame and print the dataframe
        # information.
        to_return = PTRAILDataFrame(data_set=df,
                                    latitude='lat',
                                    longitude='long',
                                    datetime='datetime',
                                    traj_id='id',
                                    rest_of_columns=[])
        print(to_return)

        # return the PTRAILDataFrame
        return to_return

    @staticmethod
    def load_seagulls():
        """
            Load the Sea-Gulls dataset into the PTRAILDataFrame and return it.

            Returns
            -------
                PTRAILDataFrame:
                    The seagulls dataset loaded into a PTrailDataFrame.
        """
        # read the CSV file from the repository using pandas.
        df = pd.read_csv(
            'https://raw.githubusercontent.com/YakshHaranwala/PTRAIL/main/examples/data/seagulls.csv'
        )

        # Load the dataset into a PTrailDataFrame and print the dataframe
        # information.
        to_return = PTRAILDataFrame(data_set=df,
                                    latitude='lat',
                                    longitude='long',
                                    datetime='DateTime',
                                    traj_id='traj_id',
                                    rest_of_columns=[])
        print(to_return)

        # return the PTRAILDataFrame
        return to_return

    @staticmethod
    def load_ships():
        """
            Load the Sea-Gulls dataset into the PTRAILDataFrame and return it.

            Returns
            -------
                PTRAILDataFrame:
                    The Ships dataset loaded into a PTrailDataFrame.
        """
        # read the CSV file from the repository using pandas.
        df = pd.read_csv(
            'https://raw.githubusercontent.com/YakshHaranwala/PTRAIL/main/examples/data/ships.csv'
        )

        # Load the dataset into a PTrailDataFrame and print the dataframe
        # information.
        to_return = PTRAILDataFrame(data_set=df.dropna(),
                                    latitude='Lat',
                                    longitude='Lon',
                                    datetime='DateTime',
                                    traj_id='VesselName')
        print(to_return)

        # return the PTRAILDataFrame
        return to_return

    @staticmethod
    def load_starkey():
        """
            Load the Starkey dataset into the PTRAILDataFrame and return it.

            Returns
            -------
                PTRAILDataFrame:
                    The Starkey dataset loaded into a PTrailDataFrame.
        """
        # read the CSV file from the repository using pandas.
        df = pd.read_csv(
            'https://raw.githubusercontent.com/YakshHaranwala/PTRAIL/main/examples/data/starkey.csv'
        )

        # Load the dataset into a PTrailDataFrame and print the dataframe
        # information.
        to_return = PTRAILDataFrame(data_set=df,
                                    latitude='lat',
                                    longitude='long',
                                    datetime='DateTime',
                                    traj_id='traj_id',
                                    rest_of_columns=[])
        print(to_return)

        # return the PTRAILDataFrame
        return to_return

    @staticmethod
    def load_starkey_habitat():
        """
            Load the Starkey dataset into a pandas dataframe and return it.

            Returns
            -------
                PTRAILDataFrame:
                    The Starkey habitat dataset.
        """
        starkey_habitat = pd.read_csv(
            'https://raw.githubusercontent.com/YakshHaranwala/PTRAIL/main/examples/data/starkey_habitat.csv'
        )
        print(f"The shape of the starkey habitat dataset: {starkey_habitat.shape}")

        return starkey_habitat



"""
    This class is an abstraction that can be used to create
    input dialog boxes for virtually any number of inputs.

    | Authors: Yaksh J Haranwala
"""
from typing import List
from PyQt5 import QtWidgets


class InputDialog(QtWidgets.QDialog):
    def __init__(self, labels: List[str], title: str, placeHolders: List[str], parent=None,):
        super().__init__(parent)
        self.setWindowTitle(title)
        self.setMinimumWidth(560)

        buttonBox = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok |
                                               QtWidgets.QDialogButtonBox.Cancel, self)
        layout = QtWidgets.QFormLayout(self)

        self.inputs = []
        for i in range(len(labels)):
            edit = QtWidgets.QLineEdit(self)
            edit.setPlaceholderText(placeHolders[i])
            self.inputs.append(edit)
            layout.addRow(labels[i], self.inputs[-1])

        layout.addWidget(buttonBox)

        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)

    def getInputs(self):
        return list(input.text() for input in self.inputs)


"""
    Contains all the default constants needed for initialization.
    All the constant are of the type string.
"""

# ------------------------------------- Column-header Constants ---------------------------------#
LAT = 'lat'
LONG = 'lon'
DateTime = 'DateTime'
TRAJECTORY_ID = 'traj_id'
OBJECT_ID = 'object_id'

MANDATORY_COLUMNS = [LAT, LONG, DateTime, TRAJECTORY_ID]

# ----------------------------------- Temporal Constants ----------------------------------------#
WEEKEND = ['Saturday', 'Sunday']

TIME_OF_DAY = [
    'Late Night',
    'Early Morning',
    'Morning',
    'Noon',
    'Evening',
    'Night'
]
# ---------------------------------- Spatial Constants -------------------------------------------#
RADIUS_OF_EARTH = 6371  # KM
PREV_DIST = 'Distance_prev_to_curr'

# ---------------------------------- Splitting Constants -----------------------------------------#
MIN_IDS = 100

# ---------------------------------- Stats Constants --------------------------------------------- #
ORDERED_COLS = [
    '10%_Distance', '25%_Distance', '50%_Distance', '75%_Distance', '90%_Distance', 'min_Distance', 'max_Distance', 'mean_Distance', 'std_Distance',
    '10%_Distance_from_start', '25%_Distance_from_start', '50%_Distance_from_start', '75%_Distance_from_start', '90%_Distance_from_start', 'min_Distance_from_start', 'max_Distance_from_start','mean_Distance_from_start', 'std_Distance_from_start',
    '10%_Speed', '25%_Speed', '50%_Speed', '75%_Speed', '90%_Speed', 'min_Speed', 'max_Speed', 'mean_Speed', 'std_Speed',
    '10%_Acceleration', '25%_Acceleration', '50%_Acceleration', '75%_Acceleration', '90%_Acceleration', 'min_Acceleration', 'max_Acceleration', 'mean_Acceleration', 'std_Acceleration',
    '10%_Jerk', '25%_Jerk', '50%_Jerk', '75%_Jerk', '90%_Jerk', 'min_Jerk', 'max_Jerk', 'mean_Jerk', 'std_Jerk',
    '10%_Bearing', '25%_Bearing', '50%_Bearing', '75%_Bearing', '90%_Bearing', 'min_Bearing', 'max_Bearing', 'mean_Bearing', 'std_Bearing',
    '10%_Bearing_Rate', '25%_Bearing_Rate', '50%_Bearing_Rate', '75%_Bearing_Rate', '90%_Bearing_Rate', 'min_Bearing_Rate','max_Bearing_Rate', 'mean_Bearing_Rate','std_Bearing_Rate',
    '10%_Rate_of_bearing_rate', '25%_Rate_of_bearing_rate', '50%_Rate_of_bearing_rate', '75%_Rate_of_bearing_rate', '90%_Rate_of_bearing_rate', 'min_Rate_of_bearing_rate','max_Rate_of_bearing_rate', 'mean_Rate_of_bearing_rate','std_Rate_of_bearing_rate',
    ]



"""
    This File contains the visualization that is a Radar Scatter plot showing how many
    number of days around each Running Water Body has an individual element spent. It
    is an interactive visualization as such the user can change the water body and check
    the distribution of animals around it.

    Warning
    -------
        The visualizations in this module are currently developed with a focus around the
        starkey.csv data as it has been developed as a side project by the developers. It
        will further be integrated into the library as a general class of visualizers in
        the time to come. Some of the visualization types may or may not work with other
        datasets.

    | Authors: Yaksh J Haranwala
"""
import pandas as pd
from IPython.core.display import display
from ipywidgets import widgets, AppLayout
from ptrail.core.TrajectoryDF import PTRAILDataFrame

from ptrail.features.temporal_features import TemporalFeatures as temp
from ptrail.preprocessing.filters import Filters as filt

import matplotlib.pyplot as plt
import seaborn as sns
import plotly.express as px

pd.options.mode.chained_assignment = None


class HydrationTrends:
    # Class variables to store datasets and the widget.
    __habitat_data = None
    __traj_data = None
    __point_dict = None
    __list = None

    @staticmethod
    def show_hydration_trends(trajectories: PTRAILDataFrame, habitat: pd.DataFrame,
                              dist_from_water: int):
        """
            Plot the interactive plotly Radar chart that shows the number of days spent
            by animals around a specific water body.

            Note
            ----
                The water bodies in the original dataset do not have any specific names.
                Hence, they are just given names such as Water-body #1, Water-body #2
                and so on.

            Parameters
            ----------
                trajectories: PTRAILDataFrame
                    The dataframe containing the trajectory data.
                habitat: pd.DataFrame
                    The dataframe containing the habitat data.
                dist_from_water: int
                    The maximum distance from the water water body that the animal should
                    be in.

            Returns
            -------
                None
        """
        # Store the datasets in the class variables.
        HydrationTrends.__habitat_data = habitat
        HydrationTrends.__traj_data = trajectories

        # First, create the date column on the trajectory dataset.
        HydrationTrends.__traj_data = temp.create_date_column(HydrationTrends.__traj_data)
        HydrationTrends.__traj_data = temp.create_time_of_day_column(HydrationTrends.__traj_data)

        # Now, filter out the moving water bodies.
        a = habitat[(habitat['EcoGener'] == 'WR')]
        water_bodies = a.loc[(a['DistEWat'] == 0)]

        # Add an extra column that has bounding boxes for all the water bodies.
        bboxes = []
        for i in range(len(water_bodies)):
            lat = water_bodies.iloc[i]['lat']
            lon = water_bodies.iloc[i]['lon']
            bboxes.append(filt.get_bounding_box_by_radius(lat, lon, dist_from_water))

        water_bodies['bbox'] = bboxes

        # Name all the water bodies and store their respective data rows in a dictionary.
        point_dict = dict()
        for i in range(len(water_bodies)):
            point_dict[f'WaterBody #{i + 1}'] = water_bodies.iloc[i]

        HydrationTrends.__point_dict = point_dict

        # Create the dropdown widget.
        HydrationTrends.__list = widgets.Dropdown(options=list(point_dict.keys()),
                                                  value='WaterBody #1',
                                                  description='WaterBody: ',
                                                  disabled=False, continuous_update=False)

        # Show the plot.
        radar = widgets.interactive_output(HydrationTrends.__plot_radar,
                                           {'body_name': HydrationTrends.__list})

        bar = widgets.interactive_output(HydrationTrends.__plot_bar,
                                         {'body_name': HydrationTrends.__list})

        output = AppLayout(header=None, center=None, footer=None,
                           left_sidebar=bar, right_sidebar=radar,
                           pane_heights=[1, 1, '480px'])

        # Display the widget and its output side-by-side.
        display(HydrationTrends.__list, output)

    @staticmethod
    def __plot_radar(body_name):
        # Get the water body data by its name.
        point = HydrationTrends.__point_dict[body_name]

        # Filter the trajectory dataset by the water body.
        dataset = filt.filter_by_bounding_box(HydrationTrends.__traj_data, point['bbox'])

        small_df = pd.DataFrame(columns=['traj_id', 'days', 'Species']).set_index('traj_id')
        ids_ = dataset.reset_index()['traj_id'].unique()
        for j in range(len(ids_)):
            mini = dataset.reset_index().loc[dataset.reset_index()['traj_id'] == ids_[j]]
            if 'D' in ids_[j]:
                small_df.loc[ids_[j]] = [pd.to_datetime(mini.Date).nunique(), 'Deer']
            elif 'E' in ids_[j]:
                small_df.loc[ids_[j]] = [pd.to_datetime(mini.Date).nunique(), 'Elk']
            else:
                small_df.loc[ids_[j]] = [pd.to_datetime(mini.Date).nunique(), 'Cattle']

        fig1 = px.scatter_polar(small_df, r='days', color='Species',
                                title=f'Number of Days Spent Around {body_name}',
                                template="ggplot2", color_discrete_sequence=px.colors.qualitative.Vivid, )

        fig1.update_layout(
            polar=dict(
                radialaxis=dict(visible=True),
            ),
            title=dict(
                font={'size': 13}
            ),
            margin=dict(l=20, r=20, t=30, b=20),
            autosize=False,
            width=540,
            height=280,
            paper_bgcolor="Gainsboro",
        )

        fig1.update_polars(
            angularaxis_tickmode='array',
            angularaxis_tickvals=[0, 90, 180, 270],
            angularaxis_ticktext=['', '', '', ''],
        )

        fig1.show()

    @staticmethod
    def __plot_bar(body_name):
        # Get the water body data by its name.
        point = HydrationTrends.__point_dict[body_name]

        # Filter the trajectory dataset by the water body.
        dataset = filt.filter_by_bounding_box(HydrationTrends.__traj_data, point['bbox'])

        new_species = []
        for i in range(len(dataset)):
            if dataset['Species'].iloc[i] == 0:
                new_species.append('Deer')
            elif dataset['Species'].iloc[i] == 1:
                new_species.append('Elk')
            else:
                new_species.append('Cattle')

        dataset['Species'] = new_species

        # Order the dataset according to the times of day.
        small_df = dataset.groupby(by=['Time_Of_Day', 'Species']).count().reset_index()
        times = ['Early Morning', 'Morning', 'Noon', 'Evening', 'Night', 'Late Night']
        small_df['Time_Of_Day'] = pd.Categorical(small_df['Time_Of_Day'],
                                                 categories=times,
                                                 ordered=True)
        small_df = small_df.sort_values(by='Time_Of_Day', ascending=True)

        # Plot the barplot from the data above.
        fig, ax = plt.subplots(figsize=(7, 5))
        sns.barplot(x='Time_Of_Day', y='lat', hue='Species', hue_order=['Deer', 'Elk', 'Cattle'], data=small_df, ax=ax)

        # Set the description of the text.
        fig.text(-0.4, 0.5,
                 f"{body_name} Description\n\n"
                 f"Pasture Name: {point['CowPast']}\n"
                 f"Coordinates: {round(point['lat'], 2), round(point['lon'], 2)} \n"
                 f"Elevation: {point['Elev']} m \n"
                 f"Canopy Cover: {point['Canopy']}%"
                 , fontdict={'size': 13})

        fig.set_facecolor('lightgray')
        ax.set_xlabel('TIME OF THE DAY')
        ax.set_ylabel('NUMBER OS TIMES VISITED')
        ax.set_title('Hydration trend of Species throughout the day')

        fig.tight_layout()



"""
    This File contains static visualizations i.e the ones that do not require the use of
    ipywidgets.

    Warning
    -------
        The visualizations in this module are currently developed with a focus around the
        starkey.csv data as it has been developed as a side project by the developers. It
        will further be integrated into the library as a general class of visualizers in
        the time to come. Some of the visualization types may or may not work with other
        datasets.

    | Authors: Yaksh J Haranwala
"""
import pandas as pd

from ptrail.core.TrajectoryDF import PTRAILDataFrame
from ptrail.features.kinematic_features import KinematicFeatures as kin
from ptrail.features.temporal_features import TemporalFeatures as temp

import plotly.express as px


class StatViz:
    @staticmethod
    def trajectory_distance_treemap(dataset: PTRAILDataFrame, path: list):
        """
            Plot a treemap of distance travelled by the moving object on a particular
            date.

            Parameters
            ----------
                dataset: PTRAILDataFrame
                    The dataframe containing all the trajectory data.
                map_date: str
                    The date for which the TreeMap is to be plotted.
                path: list
                    The hierarchy of the treemap. This is passed directly into plotly's
                    Treemap API.

            Returns
            -------
                plotly.graph_objects.Figure:
                    Treemap depicting the distance travelled.
        """
        # First obtain all the unique trajectory IDs of the dataset.
        traj_ids = dataset.reset_index()['traj_id'].unique()

        # Now, for each of the traj_id in the list above, calculate the distance
        # travelled by the moving object on that day and store it in a dictionary.
        dist_df = pd.DataFrame(columns=['traj_id', 'distance'])
        for val in traj_ids:
            try:
                distance = kin.get_distance_travelled_by_traj_id(dataframe=dataset, traj_id=val)
                duration = temp.get_traj_duration(dataframe=dataset, traj_id=val)
                dist_df.loc[val] = distance / int(duration.dt.days)

            except KeyError:
                # If the animal's trajectory is not recorded on the date given in, just skip it.
                continue

        # Drop the extra column that is acting as the traj ID and reset the index
        # and rename the index column to be traj_id.
        dist_df = dist_df.drop(columns=['traj_id']).reset_index().rename(columns={'index': 'traj_id'})

        species = []
        for i in range(len(dist_df)):
            if 'D' in dist_df.iloc[i]['traj_id']:
                species.append('Deer')
            elif 'E' in dist_df.iloc[i]['traj_id']:
                species.append('Elk')
            else:
                species.append('Cattle')

        dist_df['Species'] = species

        palette = ['#B42F32', '#DF6747', '#E3E3CD', '#878D92', '#49494D']
        # Draw the treemap using plotly.
        tree_map = px.treemap(data_frame=dist_df, values='distance', path=path,
                              color_discrete_sequence=palette, title="Average Distance Travelled Per Day")

        # Arrange the margins.
        tree_map.update_layout(margin=dict(t=50, l=25, r=25, b=25))

        # Set the color of the root of the treemap.
        tree_map.update_traces(root_color=palette[3])

        return tree_map



from .core.TrajectoryDF import PTRAILDataFrame
from .features import helper_functions
from .features import contextual_features
from .features import kinematic_features
from .features import temporal_features
from .preprocessing import filters
from .preprocessing import helpers
from .preprocessing import interpolation
from .visualization import TrajPlotter
from .visualization import HydrationTrends
from .visualization import statViz
from .visualization import InteractiveDonut

from .utilities import (
    constants,
    conversions,
    DistanceCalculator,
    exceptions,
)

__version__ = "0.7.1 Beta"



"""
    Warning
    -------
        | 1. None of the methods in this module should be used directly while performing operations on data.
        | 2. These methods are helpers for the interpolation methods in the interpolation.py
             module and hence run linearly and not in parallel which will result in slower execution time.
        | 3. All the methods in this module perform calculation on a single Trajectory ID due to which
             it will wrong results on data with multiple trajectories. Instead, use the interpolation.py
             methods for faster and reliable calculations.

    The helpers class has the functionalities that interpolate a point based
    on the given data by the user. The class contains the following 4
    interpolation calculators:

        1. Linear Interpolation
        2. Cubic Interpolation
        3. Random-Walk Interpolation
        4. Kinematic Interpolation

    Besides the interpolation helpers, there are also general utilities which
    are used in splitting up dataframes for running the code in parallel.

    | Authors: Yaksh J Haranwala, Salman Haidri
"""
import math
import os
from typing import Text, Union

import numpy as np
import pandas as pd
import datetime as dt
from hampel import hampel
from scipy.interpolate import CubicSpline

from ptrail.core.TrajectoryDF import PTRAILDataFrame
from ptrail.features.kinematic_features import KinematicFeatures as spatial
from ptrail.utilities import constants as const
from ptrail.utilities.exceptions import *


class Helpers:
    # ------------------------------------ Interpolation Helpers --------------------------------------- #
    @staticmethod
    def linear_help(dataframe: Union[pd.DataFrame, PTRAILDataFrame],
                    id_: Text, sampling_rate: float, class_label_col):
        """
            This method takes a dataframe and uses linear interpolation to determine coordinates
            of location on Datetime where the time difference between 2 consecutive points exceeds
            the user-specified sampling_rate and inserts the interpolated point those between 2 points.

            Warning
            -------
                This method should not be used for dataframes with multiple trajectory ids as it will
                yield wrong results and there might be a significant drop in performance.

            Parameters
            ----------
                dataframe: Union[pd.DataFrame, NumTrajDF]
                     The dataframe containing the original trajectory data.
                id_: Text
                    The Trajectory ID of the points in the dataframe.
                sampling_rate: float
                    The maximum time difference between 2 points greater than which
                    a point will be inserted between 2 points.

            Returns
            -------
                pandas.core.dataframe.DataFrame
                    The dataframe containing the trajectory enhanced with interpolated
                    points.

        """
        # Now, for each unique ID in the dataframe, interpolate the points.
        # Create a Series containing new times which are calculated as follows:
        #    new_time[i] = original_time[i] + sampling_rate.
        new_times = dataframe.reset_index()[const.DateTime] + pd.to_timedelta(sampling_rate, unit='seconds')

        # Now, interpolate the latitudes using numpy based on the new times calculated above.
        ip_lat = np.interp(new_times,
                           dataframe.reset_index()[const.DateTime],
                           dataframe.reset_index()[const.LAT])

        # Now, interpolate the longitudes using numpy based on the new times calculated above.
        ip_long = np.interp(new_times,
                            dataframe.reset_index()[const.DateTime],
                            dataframe.reset_index()[const.LONG])

        # Here, store the time difference between all the consecutive points in an array.
        time_deltas = dataframe.reset_index()[const.DateTime].diff().dt.total_seconds()

        # Now, for each point in the trajectory, check whether the time difference between
        # 2 consecutive points is greater than the user-specified sampling_rate, and if so then
        # insert a new point that is linearly interpolated between the 2 original points.
        for j in range(len(time_deltas)):
            if time_deltas[j] > sampling_rate:
                if class_label_col == '':
                    dataframe.loc[new_times[j - 1]] = [id_, ip_lat[j - 1], ip_long[j - 1]]
                else:
                    dataframe.loc[new_times[j - 1]] = [id_, ip_lat[j - 1], ip_long[j - 1],
                                                       dataframe[class_label_col].iloc[0]]

        return dataframe

    @staticmethod
    def cubic_help(df: Union[pd.DataFrame, PTRAILDataFrame], id_: Text,
                   sampling_rate: float, class_label_col):
        """
            This method takes a dataframe and uses cubic interpolation to determine coordinates
            of location on Datetime where the time difference between 2 consecutive points exceeds
            the user-specified sampling_rate and inserts the interpolated point those between 2 points.

            Warning
            -------
                This method should not be used for dataframes with multiple trajectory ids as it will
                yield wrong results and there might be a significant drop in performance.

            Parameters
            ----------
                df: Union[pd.DataFrame, NumTrajDF]
                     The dataframe containing the original trajectory data.
                id_: Text
                    The Trajectory ID of the points in the dataframe.
                sampling_rate: float
                    The maximum time difference between 2 points greater than which
                    a point will be inserted between 2 points.

            Returns
            -------
                pandas.core.dataframe.DataFrame
                    The dataframe containing the trajectory enhanced with interpolated
                    points.
        """
        # Create a Series containing new times which are calculated as follows:
        #    new_time[i] = original_time[i] + sampling_rate.
        new_times = df.reset_index()[const.DateTime] + pd.to_timedelta(sampling_rate, unit='seconds')

        # Now, using Scipy's Cubic spline, create a spline object for interpolation of
        # points for the dataframes which have a length greater than 3 else CubicSpline
        # doesn't execute.
        if len(df) > 3:
            # Create the x and y values for the CubicSpline function.
            # We make sure that there is a strictly increasing sequence of points.
            x = df.reset_index()[const.DateTime].sort_values().drop_duplicates()
            y = df.reset_index().iloc[x.index][[const.LAT, const.LONG]].to_numpy()

            cubic_spline = CubicSpline(x=x, y=y, extrapolate=True, bc_type='not-a-knot')
            # Now, calculate the interpolated position of the points at all the new_times
            #    calculated above.
            ip_coords = cubic_spline(new_times)

        # Here, store the time difference between all the consecutive points in an array.
        time_deltas = df.reset_index()[const.DateTime].diff().dt.total_seconds()

        # Now, for each point in the trajectory, check whether the time difference between
        # 2 consecutive points is greater than the user-specified sampling_rate, and if so then
        # insert a new point that is cubic-spline interpolated between the 2 original points.
        for j in range(len(time_deltas)):
            # If the trajectory has less than 3 points, then skip the trajectory
            # from the interpolation.
            if len(df) > 3:
                if time_deltas[j] > sampling_rate:
                    if class_label_col == '':
                        df.loc[new_times[j - 1]] = [id_, ip_coords[j - 1][0], ip_coords[j - 1][1]]
                    else:
                        df.loc[new_times[j - 1]] = [id_, ip_coords[j - 1][0], ip_coords[j - 1][1],
                                                    df[class_label_col].iloc[0]]

        return df

    @staticmethod
    def random_walk_help(dataframe: PTRAILDataFrame, id_: Text,
                         sampling_rate: float, class_label_col):
        """
            This method takes a dataframe and uses random-walk interpolation to determine coordinates
            of location on Datetime where the time difference between 2 consecutive points exceeds
            the user-specified sampling_rate and inserts the interpolated point those between 2 points.

            Warning
            -------
                This method should not be used for dataframes with multiple trajectory ids as it will
                yield wrong results and there might be a significant drop in performance.

            Parameters
            ----------
                dataframe: Union[pd.DataFrame, NumTrajDF]
                     The dataframe containing the original trajectory data.
                id_: Text
                    The Trajectory ID of the points in the dataframe.
                sampling_rate: float
                    The maximum time difference between 2 points greater than which
                    a point will be inserted between 2 points.

            Returns
            -------
                pandas.core.dataframe.DataFrame
                    The dataframe containing the trajectory enhanced with interpolated
                    points.

            References
            ----------
                Etemad, M., Soares, A., Etemad, E. et al. SWS: an unsupervised trajectory
                segmentation algorithm based on change detection with interpolation kernels.
                Geoinformatica (2020)
        """
        # Create a Series containing new times which are calculated as follows:
        #    new_time[i] = original_time[i] + sampling_rate.
        new_times = dataframe.reset_index()[const.DateTime] + pd.to_timedelta(sampling_rate, unit='seconds')

        # First, create a distance between the consecutive points of the dataframe,
        # then, calculate the mean and standard deviation of all the distances between
        # consecutive points.
        df1 = spatial.create_distance_column(dataframe)
        d_mean = (df1['Distance'].mean(skipna=True))
        d_std = (df1['Distance'].std(skipna=True))

        # Now, create a bearing between the consecutive points of the dataframe,
        # then, calculate the mean and standard deviation of all the bearings between
        # consecutive points.
        df = spatial.create_bearing_column(df1)
        b_mean = df['Bearing'].mean(skipna=True)
        b_std = df['Bearing'].std(skipna=True)

        calc_a = np.random.normal(d_mean, d_std, 1) / 1000
        calc_b = np.radians(np.random.normal(b_mean, b_std, 1))

        # print(f"Calc: {calc_a, calc_b}")
        # if d_std == 0 or b_std == 0:
        #     return dataframe
        #
        # # Here, using Scipy's truncnorm() function, create an object that gives out random
        # # values. It is to be noted that the values are restricted between latitude.min()
        # # and latitude.max().
        # d_mean = truncnorm((df1['lat'].min() - d_mean) / d_std,
        #                    (df1['lat'].max() - d_mean) / d_std,
        #                    loc=d_mean,
        #                    scale=d_std)
        #
        # # Here, using Scipy's truncnorm() function, create an object that gives out random
        # # values. It is to be noted that the values are restricted between longitude.min()
        # # and longitude.max().
        # b_mean = truncnorm((df['lon'].min() - b_mean) / b_std,
        #                    (df['lon'].max() - b_mean) / b_std,
        #                    loc=b_mean,
        #                    scale=b_std)
        #
        # # Using the 2 objects created above, generate a random value from them. The value
        # # is selected randomly from a uniformly distributed sample
        # calc_a = d_mean.rvs()
        # calc_b = math.radians(b_mean.rvs())

        dy = calc_a * np.cos(calc_b)
        dx = calc_a * np.sin(calc_b)

        # Here, store the time difference between all the consecutive points in an array.
        time_deltas = df.reset_index()[const.DateTime].diff().dt.total_seconds()

        # Look for a time diff that exceeds the sampling_rate and if one is found, calculate the
        # latitude and longitude and then append them to the dataframe at the location where
        # the threshold is crossed.
        for i in range(len(time_deltas)):
            if len(df) > 3:
                if time_deltas[i] > sampling_rate:
                    new_lat = df[const.LAT].iloc[i - 1] + (dy / const.RADIUS_OF_EARTH) * (180 / np.pi)
                    new_lon = df[const.LONG].iloc[i - 1] +\
                              (dx / const.RADIUS_OF_EARTH) * (180 / np.pi) / np.cos(df[const.LAT].iloc[i - 1] * np.pi / 180)
                    if class_label_col == '':
                        dataframe.loc[new_times[i - 1]] = [id_, new_lat[0], new_lon[0]]
                    else:
                        dataframe.loc[new_times[i - 1]] = [id_, new_lat[0], new_lon[0],
                                                           dataframe[class_label_col].iloc[0]]


        # Return the new dataframe
        return dataframe

    @staticmethod
    def kinematic_help(dataframe: Union[pd.DataFrame, PTRAILDataFrame], id_: Text,
                       sampling_rate: float, class_label_col):
        """
            This method takes a dataframe and uses kinematic interpolation to determine coordinates
            of location on Datetime where the time difference between 2 consecutive points exceeds
            the user-specified sampling_rate and inserts the interpolated point those between 2 points.

            Warning
            -------
                This method should not be used for dataframes with multiple trajectory ids as it will
                yield wrong results and there might be a significant drop in performance.

            Parameters
            ----------
                dataframe: Union[pd.DataFrame, NumTrajDF]
                     The dataframe containing the original trajectory data.
                id_: Text
                    The Trajectory ID of the points in the dataframe.
                sampling_rate: float
                    The maximum time difference between 2 points greater than which
                    a point will be inserted between 2 points.

            Returns
            -------
                pandas.core.dataframe.DataFrame
                    The dataframe containing the trajectory enhanced with interpolated
                    points.

            References
            ----------
                Nogueira, T.O., "kinematic_interpolation.py", (2016), GitHub repository,
                https://gist.github.com/talespaiva/128980e3608f9bc5083b.js
        """
        # Create a Series containing new times which are calculated as follows:
        #    new_time[i] = original_time[i] + sampling_rate.
        new_times = dataframe.reset_index()[const.DateTime] + pd.to_timedelta(sampling_rate, unit='seconds')

        # Here, store the time difference between all the consecutive points in an array.
        time_deltas = dataframe.reset_index()[const.DateTime].diff().dt.total_seconds()
        lat_diff = dataframe.reset_index()[const.LAT].diff()
        lon_diff = dataframe.reset_index()[const.LONG].diff()

        lat_velocity = lat_diff / time_deltas
        lon_velocity = lon_diff / time_deltas

        lat = list(dataframe.reset_index()[const.LAT].values)
        lon = list(dataframe.reset_index()[const.LONG].values)

        # Look for a time diff that exceeds the sampling_rate and if one is found, calculate the
        # latitude and longitude and then append them to the dataframe at the location where
        # the threshold is crossed.
        for i in range(len(time_deltas)):
            if time_deltas[i] > sampling_rate and not np.isnan(lat_velocity[i - 1]):
                ax = np.array([[(time_deltas[i] ** 2) / 2, (time_deltas[i] ** 3) / 6],
                               [float(time_deltas[i]), (time_deltas[i] ** 2) / 2]])
                bx = [lat[i] - lat[i - 1] - lat_velocity[i - 1] * time_deltas[i], lat_velocity[i] - lat_velocity[i - 1]]
                coef_x = np.linalg.solve(ax, bx)

                ay = ax
                by = [lon[i] - lon[i - 1] - lon_velocity[i - 1] * time_deltas[i], lon_velocity[i] - lon_velocity[i - 1]]
                coef_y = np.linalg.solve(ay, by)

                td = new_times[i - 1].timestamp() / 10e9
                x = Helpers._pos(t=td, x1=lat[i - 1], v1=lat_velocity[i - 1], b=coef_x[0], c=coef_x[1])
                y = Helpers._pos(t=td, x1=lon[i - 1], v1=lon_velocity[i - 1], b=coef_y[0], c=coef_y[1])

                if class_label_col == '':
                    dataframe.loc[new_times[i - 1]] = [id_, x, y]
                else:
                    dataframe.loc[new_times[i - 1]] = [id_, x, y, dataframe[class_label_col].iloc[0]]

        return dataframe

    @staticmethod
    def hampel_help(df, column_name):
        """
            This function is the helper function for the hampel_outlier_detection()
            function present in the filters module. The purpose of the function is to
            run the hampel filter on a single trajectory ID, remove the outliers
            and return the smaller dataframe.

            Warning
            -------
                This function should not be used directly as it will result in a
                slower execution of the function and might result in removal of
                points that are actually not outliers.

            Warning
            -------
                Do not use Hampel filter outlier detection and try to detect outliers
                with DateTime as it will raise a NotImplementedError as it has not been
                implemented yet by the original author of the Hampel filter.

            Parameters
            ----------
                df: PTRAILDataFrame/pd.core.dataframe.DataFrame
                    The dataframe which the outliers are to be removed
                column_name: Text
                    The column based on which the outliers are to be removed.

            Returns
            -------
                pd.core.dataframe.DataFrame
                    The dataframe where the outlier points are removed.

        """
        try:
            # First, extract the column from the dataframe and then obtain the
            # outlier indices which are to be removed.
            col = df[column_name]
            outlier_indices = hampel(col)

            # Now, drop the indices given out by the hampel filter.
            to_return = df.drop(df.index[outlier_indices])
            #

            return to_return

        except KeyError:
            raise MissingColumnsException(f"The column {column_name} does not exist in the dataset."
                                          f"Please check the column name and try again.")

    @staticmethod
    def _pos(t, x1, v1, b, c):
        return x1 + v1 * t + (t ** 2) * b / 2 + (t ** 3) * c / 6

    # ------------------------------------------ Statistics Helpers ----------------------------------- #
    @staticmethod
    def split_traj_helper(df, num_days):
        # First, create the date column and get all the unique traj_ids
        # in the dataframe.
        df['Date'] = df[const.DateTime].dt.date
        ids_ = list(df.traj_id.value_counts().keys())

        df_chunks = []
        for i in range(len(ids_)):
            small_df = df.reset_index().loc[df.reset_index()[const.TRAJECTORY_ID] == ids_[i]]
            df_chunks.append(small_df)

        # Now, iterate over the entire dataframe and then segment
        # the trajectories by num_days each.
        results = []
        for i in range(len(ids_)):
            # Take the traj_df of a single Trajectory out from the
            # list of chunks and find their max and min timestamps.
            traj = df_chunks[i]
            t_max = traj[const.DateTime].max()
            t_min = traj[const.DateTime].min()

            # For iteration purposes, set t_1 to min and t_2 to
            # t_1 + num_days days.
            t_1 = t_min
            t_2 = t_1 + dt.timedelta(days=num_days)
            seg_id = 1

            # Now, segment the trajectories into smaller segments
            # wherein each segment contains the points of a span
            # of num_days days only.
            while t_2 < t_max:
                if t_2 < t_max:
                    seg = Helpers.filt_df_by_date(traj,
                                                  start_date=t_1.strftime('%Y-%m-%d'),
                                                  end_date=t_max.strftime('%Y-%m-%d'))
                    # Once filtered, assign the segment with a segment ID.
                    seg['seg_id'] = seg_id

                    # Increment the segment id, t_1 and t_2 values by
                    # 1, num_days days each respectively to continue the iteration.
                    t_1 += dt.timedelta(days=num_days)
                    t_2 += dt.timedelta(days=num_days)

                    if len(seg) > 0:
                        seg_id += 1

                    results.append(seg.drop(columns=['index', 'level_0']))

                # If, t_2 is greater than the max time present in the
                # trajectory, then assign t_2 = max and proceed
                # further with segmentation.
                elif t_2 >= t_max:
                    seg = Helpers.filt_df_by_date(traj,
                                                  start_date=t_1.strftime('%Y-%m-%d'),
                                                  end_date=t_max.strftime('%Y-%m-%d'))
                    # Once filtered, assign the segment with a segment ID.
                    seg['seg_id'] = seg_id

                    # Increment the segment id, t_1 and t_2 values by
                    # 1, num_days each respectively to continue the iteration.
                    t_1 += dt.timedelta(days=num_days)
                    t_2 += dt.timedelta(days=num_days)

                    if len(seg) > 0:
                        seg_id += 1

                    results.append(seg.drop(columns=['index', 'level_0']))
                    seg_id += 1

        # Finally, concat the dataframes, set the index as
        # [traj_id, seg_id, DateTime].
        return pd.concat(results).reset_index().set_index(['traj_id', 'seg_id', 'DateTime']).sort_values(by=['traj_id',
                                                                                                             'seg_id'])

    @staticmethod
    def filt_df_by_date(dataframe, start_date, end_date):
        # Convert the user-given string dates to pandas datetime format.
        start_date = pd.to_datetime(start_date) if start_date is not None else None
        end_date = pd.to_datetime(end_date) if end_date is not None else None

        # Case-1: No start and end date are give. Hence just return the original dataframe.
        if start_date is None and end_date is None:
            filtered_df = dataframe

        # Case-2: No start_date is given. Hence, return all the points upto and including
        #         the points on the end date.
        elif start_date is None and end_date is not None:
            filt = dataframe['Date'] <= end_date
            filtered_df = dataframe.loc[filt]

        # Case-3: No end date is given. Hence, return all the point after and including the
        #         points on the start date.
        elif start_date is not None and end_date is None:
            filt = dataframe['Date'] >= start_date
            filtered_df = dataframe.loc[filt]

        # Case-4: Both the start date and end date are given. Hence, return the points between
        #         and including the points on start and end date.
        else:
            if end_date < start_date:
                raise ValueError(f"End Date should be later than Start Date.")
            else:
                filt = np.logical_and(dataframe['Date'] >= start_date, dataframe['Date'] <= end_date)
                filtered_df = dataframe.loc[filt].reset_index()

        # Convert the smaller dataframe back to PTRAILDataFrame and return it.
        return filtered_df

    @staticmethod
    def stats_helper(df, target_col_name, segmented):
        """
            Generate the stats of the kinematic features present in the Dataframe.

            Parameters
            ----------
                df: pandas.core.dataframe.DataFrame
                    The dataframe containing the trajectory data and their features.
                target_col_name: str
                    This is the 'y' value that is used for ML tasks, this is
                    asked to append the species back at the end.
                segmented: Optional[bool]
                    Indicate whether the trajectory has segments or not.

            Returns
            -------
                pd.core.dataframe.DataFrame:
                    A dataframe containing the stats of the given trajectory.

        """
        # Grab the columns that we need.
        if not segmented:
            new_df = df.reset_index()[['traj_id', 'Distance', 'Distance_from_start', 'Speed',
                                       'Acceleration', 'Jerk', 'Bearing', 'Bearing_Rate',
                                       'Rate_of_bearing_rate']]
            # Generate the stats along with the needed percentiles and arrange the dataframe
            # properly.
            stats = new_df.reset_index(drop=True).describe(percentiles=[0.1, 0.25, 0.5, 0.75, 0.9]).transpose()

            # Assign the traj_id column.
            stats['traj_id'] = new_df['traj_id'].iloc[0]
            stats = stats[['traj_id', 'mean', 'std', 'min', '10%',
                           '25%', '50%', '75%', '90%', 'max']]
            stats[target_col_name] = df[target_col_name].iloc[0]

            return stats.reset_index().rename(
                columns={'index': 'Columns'}).reset_index(drop=True).set_index(['traj_id', 'Columns'])
        else:
            seg_id = df['seg_id'].iloc[0]
            new_df = df.reset_index()[['traj_id', 'Distance', 'Distance_from_start', 'Speed',
                                       'Acceleration', 'Jerk', 'Bearing', 'Bearing_Rate',
                                       'Rate_of_bearing_rate']]

            # Generate the stats along with the needed percentiles and arrange the dataframe
            # properly.
            stats = new_df.reset_index(drop=True).describe(percentiles=[0.1, 0.25, 0.5, 0.75, 0.9]).transpose()

            # Assign the traj_id column.
            stats['traj_id'] = new_df['traj_id'].iloc[0]
            stats['seg_id'] = seg_id
            stats = stats[['traj_id', 'seg_id', 'mean', 'std', 'min', '10%',
                           '25%', '50%', '75%', '90%', 'max']]
            stats[target_col_name] = df[target_col_name].iloc[0]
            stats = stats.loc[:, ~stats.columns.duplicated()]

            to_return = stats.reset_index().rename(
                columns={'index': 'Columns'}).reset_index(drop=True).set_index(['traj_id', 'seg_id', 'Columns'])
            return to_return

    # -------------------------------------- General Utilities ---------------------------------- #
    @staticmethod
    def _get_partition_size(size):
        """
            Takes number of ids and makes use of a formula that gives a factor to makes set of ids
            according to the number of processors available to work with.

            Parameters
            ----------
                size: int
                    The total number of trajectory IDs in the dataset.

            Returns
            -------
                int
                    The factor by which the datasets are to be split.
        """
        # Based on the Operating system, get the number of CPUs available for
        # multiprocessing.
        num = os.cpu_count()
        NUM_CPU = math.ceil((num * 2) / 3)

        # Integer divide the total number of Trajectory IDs by the number of available CPUs
        # The factor of 1 is added to avoid errors when the integer division yields a 0.
        factor = (size // NUM_CPU) + 1

        # Return the factor if it is less than 100 otherwise return 100.
        # This factor hence is capped at 100.
        return factor if factor < 100 else 100

    @staticmethod
    def _df_split_helper(dataframe):
        """
            This is the helper function for splitting up dataframes into smaller chunks.
            This function is widely used for main functions to help split the original
            dataframe into smaller chunks based on a fixed range of IDs. This function
            splits the dataframes based on a predetermined number, stores them in a list
            and returns it.
            NOTE: The dataframe is split based on the number of CPU cores available for.
                  For more info, take a look at the documentation of the get_partition_size()
                  function.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe that is to be split.

            Returns
            -------
                list:
                    The list containing smaller dataframe chunks.
        """
        # First, create a list containing all the ids of the data and then further divide that
        # list items and split it into sub-lists of ids equal to split_factor.
        ids_ = list(dataframe.reset_index().traj_id.value_counts().keys())

        # Get the ideal number of IDs by which the dataframe is to be split.
        split_factor = Helpers._get_partition_size(len(ids_))
        ids_ = [ids_[i: i + split_factor] for i in range(0, len(ids_), split_factor)]

        # Now split the dataframes based on set of Trajectory ids.
        # As of now, each smaller chunk is supposed to have data of 100
        # trajectory IDs max
        df_chunks = [dataframe.loc[dataframe[const.TRAJECTORY_ID].isin(ids_[i])]
                     for i in range(len(ids_))]
        return df_chunks



"""
    The statistics module has several functionalities that calculate kinematic
    statistics of the trajectory, split trajectories, pivot dataframes etc.
    The main purpose of this module is to get the dataframe ready for Machine
    Learning tasks such as clustering, calssification etc.

    | Author: Yaksh J Haranwala

"""
import itertools
from typing import Optional

import pandas as pd

from ptrail.core.TrajectoryDF import PTRAILDataFrame
from ptrail.features.kinematic_features import KinematicFeatures
from ptrail.preprocessing.helpers import Helpers as helpers
import ptrail.utilities.constants as const
from math import ceil

import multiprocessing
import os

num = os.cpu_count()
NUM_CPU = ceil((num * 2) / 3)


class Statistics:
    @staticmethod
    def segment_traj_by_days(dataframe: PTRAILDataFrame, num_days):
        """
            Given a dataframe containing trajectory data, segment all
            the trajectories by each week.

            Parameters
            ----------
                df: PTRAILDataFrame
                    The dataframe containing trajectory data.
                num_days: int
                    The number of days that each segment is supposed to have.

            Returns
            -------
                pandas.core.dataframe.DataFrame:
                    The dataframe containing segmented trajectories
                    with a new column added called segment_id
        """
        # splitting the dataframe according to trajectory ids
        df_chunks = helpers._df_split_helper(dataframe=dataframe.reset_index())

        # Here, create 2/3rds number of processes as there are in the system. Some CPUs are
        # kept free at all times in order to not block up the system.
        # (Note: The blocking of system is mostly prevalent in Windows and does not happen very often
        # in Linux. However, out of caution some CPUs are kept free regardless of the system.)
        pool = multiprocessing.Pool(NUM_CPU)
        results = pool.starmap(helpers.split_traj_helper, zip(df_chunks, itertools.repeat(num_days)))
        pool.close()
        pool.join()

        to_return = pd.concat(results).reset_index().set_index(['traj_id', 'seg_id', 'DateTime'])

        return to_return.drop(columns=['index'])

    @staticmethod
    def generate_kinematic_stats(dataframe: PTRAILDataFrame, target_col_name: str, segmented: Optional[bool] = False):
        """
            Generate the statistics of kinematic features for each unique trajectory in
            the dataframe.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe containing the trajectory data.
                target_col_name: str
                    This is the 'y' value that is used for ML tasks, this is
                    asked to append the target_col back at the end.
                segmented: Optional[bool]
                    Indicate whether the trajectory has segments or not.

            Returns
            -------
                pandas.core.dataframe.DataFrame:
                    A pandas dataframe containing stats for all kinematic features for
                    each unique trajectory in the dataframe.
        """
        # Generate kinematic features on the entire dataframe.
        ptdf = KinematicFeatures.generate_kinematic_features(dataframe)

        # Then, lets break down the entire dataframe into pieces containing data of
        # 1 trajectory in each piece.
        ids_ = list(dataframe.reset_index().traj_id.value_counts().keys())
        df_chunks = []

        if segmented:
            segments = ptdf.reset_index()['seg_id'].unique()
            for i in range(len(ids_)):
                for seg in segments:
                    df = ptdf.reset_index().loc[(ptdf.reset_index()['traj_id'] == ids_[i])
                                                & (ptdf.reset_index()['seg_id'] == seg)]
                    if len(df) > 0:
                        df_chunks.append(df)
                    else:
                        continue
        else:
            for i in range(len(ids_)):
                small_df = ptdf.reset_index().loc[ptdf.reset_index()[const.TRAJECTORY_ID] == ids_[i]]
                df_chunks.append(small_df)

        # Here, create 2/3rds number of processes as there are in the system. Some CPUs are
        # kept free at all times in order to not block up the system.
        # (Note: The blocking of system is mostly prevalent in Windows and does not happen very often
        # in Linux. However, out of caution some CPUs are kept free regardless of the system.)
        mp_pool = multiprocessing.Pool(NUM_CPU)
        results = mp_pool.starmap(helpers.stats_helper, zip(df_chunks,
                                                            itertools.repeat(target_col_name),
                                                            itertools.repeat(segmented)))
        mp_pool.close()
        mp_pool.join()

        return pd.concat(results)

    @staticmethod
    def pivot_stats_df(dataframe, target_col_name: str, segmented: Optional[bool] = False):
        """
            Given a dataframe with stats present in it, melt the dataframe to make it
            ready for ML tasks. This is specifically for melting the type of dataframe
            generated by the generate_kinematic_stats() function of the kinematic_features
            module.

            Check the kinematic_features module for further details about the dataframe
            expected.

            Parameters
            ----------
                dataframe: pd.core.dataframe.DataFrame
                    The dataframe containing stats.
                target_col_name: str
                    This is the 'y' value that is used for ML tasks, this is
                    asked to append the target_col back at the end.
                segmented: Optional[bool]
                    Indicate whether the trajectory has segments or not.

            Returns
            -------
                pd.core.dataframe.DataFrame:
                    The dataframe above which is pivoted and has rows converted to columns.
        """
        # Get all the unique trajectory IDs.
        ids_ = list(dataframe.index.get_level_values('traj_id').unique())

        final_chunks = []
        if not segmented:
            for val in ids_:
                # separated the data for each trajectory id.
                small = dataframe.loc[dataframe.index.get_level_values('traj_id') == val].reset_index().set_index(
                    'traj_id')

                # Get the target value out and drop the target column.
                target = small[target_col_name].iloc[0]
                small = small.drop(columns=[target_col_name])

                # Pivot the table now and adjust the column names.
                pivoted = small.reset_index().pivot_table(index='traj_id', columns='Columns')
                pivoted.columns = pivoted.columns.map('_'.join).str.strip('|')

                # Assign the target column again.
                pivoted[target_col_name] = target
                final_chunks.append(pivoted)
        else:
            for val in ids_:
                segments = dataframe.index.get_level_values('seg_id').unique()
                for seg in segments:
                    small = dataframe.loc[(dataframe.index.get_level_values('traj_id') == val) &
                                          (dataframe.index.get_level_values('seg_id') == seg)].reset_index().set_index(['traj_id', 'seg_id'])

                    if len(small) <= 0:
                        continue

                    # Get the target value out and drop the target column.
                    target = small[target_col_name].iloc[0]
                    small = small.drop(columns=[target_col_name])

                    # Pivot the table now and adjust the column names.
                    pivoted = small.reset_index().pivot_table(index=['traj_id', 'seg_id'], columns='Columns')
                    pivoted.columns = pivoted.columns.map('_'.join).str.strip('|')

                    # Assign the target column again.
                    pivoted[target_col_name] = target
                    final_chunks.append(pivoted)

        # Concatenate the smaller chunks and reorder the columns.
        to_return = pd.concat(final_chunks)

        # Store the correct order of the columns to a variable and add the name
        # of the target column to the end of it.
        cols = const.ORDERED_COLS
        cols.append(target_col_name)

        # Reorder the final DF, drop duplicated columns and return it.
        to_return = to_return[cols]
        to_return = to_return.loc[:, ~to_return.columns.duplicated()]
        return to_return



"""
    | 1. The temporal_features module contains all the features of the library
         that calculates several features based on the DateTime provided in
         the data.
    | 2. It is to be noted that most of the functions in this module
         calculate the features and then add the results to an entirely new
         column with a new column header.
    | 3. It is to be also noted that a lot of these features are inspired from
         the PyMove library and we are crediting the PyMove creators with them.

    | Authors: Yaksh J Haranwala, Salman Haidri

    References
    ----------
        Arina De Jesus Amador Monteiro Sanches. “Uma Arquitetura E Imple-menta ̧c ̃ao Do M ́odulo De
        Pr ́e-processamento Para Biblioteca Pymove”.Bachelor’s thesis. Universidade Federal Do Cear ́a, 2019
"""
import itertools
import multiprocessing
from typing import Optional, Text

import numpy as np
import pandas as pd

from ptrail.core.TrajectoryDF import PTRAILDataFrame
from ptrail.features.helper_functions import Helpers as helpers
from ptrail.utilities import constants as const


class TemporalFeatures:
    @staticmethod
    def create_date_column(dataframe: PTRAILDataFrame):
        """
            From the DateTime column already present in the data, extract only the date
            and then add another column containing just the date.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The PTRAILDataFrame Dataframe on which the creation of the time column is to be done.

            Returns
            -------
                PTRAILDataFrame:
                    The dataframe containing the resultant Date column.

        """
        df = dataframe.reset_index()

        # From the DateTime value extract the dates and store them in Date column
        df['Date'] = df[const.DateTime].dt.date

        # Return the dataframe by converting it to PTRAILDataFrame
        return PTRAILDataFrame(df.reset_index(drop=True),
                               const.LAT, const.LONG, const.DateTime, const.TRAJECTORY_ID)

    @staticmethod
    def create_time_column(dataframe: PTRAILDataFrame):
        """
            From the DateTime column already present in the data, extract only the time
            and then add another column containing just the time.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The PTRAILDataFrame Dataframe on which the creation of the time column is to be done.

            Returns
            -------
                PTRAILDataFrame
                    The dataframe containing the resultant Time column.

        """
        dataframe = dataframe.reset_index()

        # From the DateTime column extract the time and store them in the Time column
        dataframe['Time'] = dataframe[const.DateTime].dt.time

        # Return the dataframe by converting it into PTRAILDataFrame type
        return PTRAILDataFrame(dataframe, const.LAT, const.LONG, const.DateTime, const.TRAJECTORY_ID)

    @staticmethod
    def create_day_of_week_column(dataframe: PTRAILDataFrame):
        """
            Create a column called Day_Of_Week which contains the day of the week
            on which the trajectory point is recorded. This is calculated on the basis
            of timestamp recorded in the data.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe containing the entire data on which the operation is to be performed

            Returns
            -------
                PTRAILDataFrame
                    The dataframe containing the resultant Day_of_week column.
        """
        dataframe = dataframe.reset_index()

        # From the DateTime column extract the time and store them in the Time column
        dataframe['Day_Of_Week'] = dataframe[const.DateTime].dt.day_name()

        # Return the dataframe by converting it into PTRAILDataFrame type
        return PTRAILDataFrame(dataframe, const.LAT, const.LONG, const.DateTime, const.TRAJECTORY_ID)

    @staticmethod
    def create_weekend_indicator_column(dataframe: PTRAILDataFrame):
        """
            Create a column called Weekend which indicates whether the point data is collected
            on either a Saturday or a Sunday.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe on which the operation is to be performed.

            Returns
            -------
                PTRAILDataFrame
                    The dataframe containing the resultant Weekend column.

            References
            ----------
                Arina De Jesus Amador Monteiro Sanches. 'Uma Arquitetura E Imple-menta ̧c ̃ao
                Do M ́odulo De Pr ́e-processamento Para Biblioteca Pymove'.Bachelor’s thesis.
                Universidade Federal Do Cear ́a, 2019.

        """
        # Check if Day_of_Week column is present in the dataframe
        if 'Day_Of_Week' in dataframe.columns:
            dataframe = dataframe.reset_index()
            # store the weekend days in a series
            fd = np.logical_or(dataframe['Day_Of_Week'] == const.WEEKEND[0],
                               dataframe['Day_Of_Week'] == const.WEEKEND[1])
            # store the index of the weekends
            index_fd = dataframe[fd].index
            # initialize the Weekend column with False and then update all the indexes which
            # indicates that it's a weekend
            dataframe['Weekend'] = False
            dataframe.at[index_fd, 'Weekend'] = True
            # Return the dataframe by converting it into PTRAILDataFrame
        else:
            # If the dataframe does not contain the Day_Of_Week column already,
            # then call the method and create the column.
            dataframe = TemporalFeatures.create_day_of_week_column(dataframe)
            dataframe = dataframe.reset_index()
            # store the weekend days in a series
            fd = np.logical_or(dataframe['Day_Of_Week'] == const.WEEKEND[0],
                               dataframe['Day_Of_Week'] == const.WEEKEND[1])
            # store the index of the weekends
            index_fd = dataframe[fd].index
            # initialize the Weekend column with False and then update all the indexes which
            # indicates that it's a weekend
            dataframe['Weekend'] = False
            dataframe.at[index_fd, 'Weekend'] = True
            # Return the dataframe by converting it into PTRAILDataFrame

        return PTRAILDataFrame(dataframe, const.LAT, const.LONG, const.DateTime, const.TRAJECTORY_ID)

    @staticmethod
    def create_time_of_day_column(dataframe: PTRAILDataFrame):
        """
            Create a Time_Of_Day column in the dataframe using parallelization which indicates at what time of the
            day was the point data captured.
            Note: The divisions of the day based on the time are provided in the utilities.constants module.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe on which the calculation is to be done.

            Returns
            -------
                PTRAILDataFrame
                    The dataframe containing the resultant Time_Of_Day column.

            References
            ----------
                Arina De Jesus Amador Monteiro Sanches. 'Uma Arquitetura E Imple-menta ̧c ̃ao
                Do M ́odulo De Pr ́e-processamento Para Biblioteca Pymove'.Bachelor’s thesis.
                Universidade Federal Do Cear ́a, 2019.

        """
        dataframe = dataframe.reset_index()
        # Extract the hours from the Datetime column and then create a list of conditions for the
        # different periods of time
        hours = dataframe[const.DateTime].dt.hour
        conditions = [
            (hours >= 0) & (hours <= 4),
            (hours > 4) & (hours <= 8),
            (hours > 8) & (hours <= 12),
            (hours > 12) & (hours <= 16),
            (hours > 16) & (hours <= 20),
            (hours > 20) & (hours <= 24)
        ]
        # Map the conditions to the different periods of  the day
        dataframe['Time_Of_Day'] = np.select(conditions, const.TIME_OF_DAY)
        return PTRAILDataFrame(dataframe, const.LAT, const.LONG, const.DateTime, const.TRAJECTORY_ID)

    @staticmethod
    def get_traj_duration(dataframe: PTRAILDataFrame, traj_id: Optional[Text] = None):
        """
            Accessor method for the duration of a trajectory specified by the user.

            Note
            ----
                If no trajectory ID is given by the user, then the duration of each
                unique trajectory is calculated.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe containing the resultant column if inplace is True.
                traj_id: Optional[Text]
                    The trajectory id for which the duration is required.

            Returns
            -------
                pandas.TimeDelta:
                    The trajectory duration.
                pandas.core.dataframe.DataFrame:
                    The dataframe containing the duration of all trajectories in the dataset.
        """
        dataframe = dataframe.reset_index()
        if traj_id is None:
            ids_ = dataframe[const.TRAJECTORY_ID].value_counts(ascending=True).keys().to_list()

            split_factor = helpers._get_partition_size(len(ids_))
            ids_ = [ids_[i: i + split_factor] for i in range(0, len(ids_), split_factor)]

            mp_pool = multiprocessing.Pool(len(ids_))
            results = mp_pool.starmap(helpers.traj_duration_helper, zip(itertools.repeat(dataframe), ids_))
            mp_pool.close()
            mp_pool.join()

            results = pd.concat(results).sort_values(const.TRAJECTORY_ID)
            return results
        else:
            small = dataframe.loc[dataframe[const.TRAJECTORY_ID] == traj_id, [const.DateTime]]
            if len(small) == 0:
                return f"No {traj_id} exists in the given data. Please try again."
            else:
                return small.max() - small.min()

    @staticmethod
    def get_start_time(dataframe: PTRAILDataFrame, traj_id: Optional[Text] = None):
        """
            Get the starting time of the trajectory.

            Note
            ----
                If the trajectory ID is not specified by the user, then by default,
                the starting times of all the trajectory IDs in the data are
                returned.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe on which the operations are to be performed.
                traj_id: Optional[Text]
                    The trajectory for which the start time is required.

            Returns
            -------
                pandas.DateTime:
                    The start time of a single trajectory.
                pandas.core.dataframe.DataFrame
                    Pandas dataframe containing the start time of all the trajectories
                    present in the data when the user hasn't asked for a particular
                    trajectory's start time.
        """
        dataframe = dataframe.reset_index()
        if traj_id is None:
            # First, create a list containing all the ids of the data and then further divide that
            # list items and split it into sub-lists of 100 ids each if there are more than 100 ids.
            ids_ = dataframe[const.TRAJECTORY_ID].value_counts(ascending=True).keys().to_list()

            # Get the ideal number of IDs by which the dataframe is to be split.
            split_factor = helpers._get_partition_size(len(ids_))
            ids_ = [ids_[i: i + split_factor] for i in range(0, len(ids_), split_factor)]

            # Now, create a multiprocessing pool and then run processes in parallel
            # which calculate the start times for a smaller set of IDs only.
            mp_pool = multiprocessing.Pool(len(ids_))
            results = mp_pool.starmap(helpers.start_time_helper, zip(itertools.repeat(dataframe), ids_))
            mp_pool.close()
            mp_pool.join()

            # Concatenate all the smaller dataframes and return the answer.
            results = pd.concat(results).sort_values(const.TRAJECTORY_ID)
            return results
        else:
            filt = dataframe.loc[dataframe[const.TRAJECTORY_ID] == traj_id]
            filt_two = filt.loc[filt[const.DateTime] == filt[const.DateTime].min()]
            return filt_two[const.DateTime].iloc[0]

    @staticmethod
    def get_end_time(dataframe: PTRAILDataFrame, traj_id: Optional[Text] = None):
        """
            Get the ending time of the trajectory.

            Note
            ----
                If the trajectory ID is not specified by the user, then by default,
                the ending times of all the trajectory IDs in the data are
                returned.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe on which the operations are to be performed.
                traj_id: Optional[Text]
                    The trajectory for which the end time is required.

            Returns
            -------
                pandas.DateTime:
                    The end time of a single trajectory.
                pandas.core.dataframe.DataFrame
                    Pandas dataframe containing the end time of all the trajectories
                    present in the data when the user hasn't asked for a particular
                    trajectory's end time.
        """
        dataframe = dataframe.reset_index()
        if traj_id is None:
            # First, create a list containing all the ids of the data and then further divide that
            # list items and split it into sub-lists of 100 ids each if there are more than 100 ids.
            ids_ = dataframe[const.TRAJECTORY_ID].value_counts(ascending=True).keys().to_list()

            # Get the ideal number of IDs by which the dataframe is to be split.
            split_factor = helpers._get_partition_size(len(ids_))
            ids_ = [ids_[i: i + split_factor] for i in range(0, len(ids_), split_factor)]

            # Now, create a multiprocessing pool and then run processes in parallel
            # which calculate the end times for a smaller set of IDs only.
            mp_pool = multiprocessing.Pool(len(ids_))
            results = mp_pool.starmap(helpers.end_time_helper, zip(itertools.repeat(dataframe), ids_))
            mp_pool.close()
            mp_pool.join()

            # Concatenate all the smaller dataframes and return the answer.
            results = pd.concat(results).sort_values(const.TRAJECTORY_ID)
            return results
        else:
            filt = dataframe.loc[dataframe[const.TRAJECTORY_ID] == traj_id]
            filt_two = filt.loc[filt[const.DateTime] == filt[const.DateTime].max()]
            return filt_two[const.DateTime].iloc[0]

    @staticmethod
    def generate_temporal_features(dataframe: PTRAILDataFrame):
        """
            Generate all the temporal features with a single call of this function.
            
            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe on which the features are to be generated.
            
            Returns
            -------
                PTRAILDataFrame:
                    The dataframe enriched with Temporal Features.
        """
        to_return = TemporalFeatures.create_date_column(dataframe)
        to_return = TemporalFeatures.create_time_column(to_return)
        to_return = TemporalFeatures.create_day_of_week_column(to_return)
        to_return = TemporalFeatures.create_weekend_indicator_column(to_return)
        to_return = TemporalFeatures.create_time_of_day_column(to_return)

        return to_return


"""
    This module contains the design of PTRAIL's GUI module.
    It is to be noted that this class does not handle the functionalities,
    it is rather handled by the handler class.

    | Author: Yaksh J Haranwala
"""
import sys

from PyQt5 import QtCore, QtWidgets, QtGui, QtWebEngineWidgets
from PyQt5.QtWidgets import QSizePolicy

from ptrail.GUI.handler import GuiHandler


# TODO: The application crashes when we cancel file upload and try to upload a different file.
class Ui_MainWindow(QtWidgets.QMainWindow):
    def __init__(self, OuterWindow):
        super(Ui_MainWindow, self).__init__()
        # Main Window variables.
        self.statusBar = None
        self.vlayout = None
        self.centralwidget = None

        # MenuBar variables
        self.VersionInfoButton = None
        self.QuitButton = None
        self.OpenButton = None
        self.SaveButton = None
        self.menuAbout = None
        self.FileMenu = None
        self.MenuBar = None

        # MapPane variables.
        self.maplayoutmanager = None
        self.MapPane = None
        self.map = None
        self.open_btn = None

        # CommandPalette variables
        self.cmdlayoutmanager = None
        self.CommandPalette = None
        self.featureType = None
        self.featureListWidget = None
        self.dropColumnWidget = None
        self.runStatsBtn = None
        self.dropColumnBtn = None

        # StatsPane variables.
        self.statslayoutmanager = None
        self.StatsPane = None
        self.selectStatDropdown = None
        self.figure = None
        self.canvas = None

        # DFPane variable.
        self.dflayoutmanager = None
        self.DFPane = None
        self.OuterWindow = OuterWindow
        self.exportBtn = None
        self.dfController = None
        self.dfView = None

        # The GUI conductor.
        self.handler = None
        self.setupUi(OuterWindow)

    def setupUi(self, OuterWindow):
        """
            Set the main window of the GUI up and start the application.

            Parameters
            ----------
                OuterWindow: PyQt5.QtWidgets.QOuterWindow'
        """
        OuterWindow.setObjectName("OuterWindow")
        OuterWindow.resize(1125, 776)

        self.centralwidget = QtWidgets.QWidget(OuterWindow)
        self.vlayout = QtWidgets.QGridLayout(self.centralwidget)
        self.centralwidget.setObjectName("centralwidget")

        self.setup_menubar()
        self.setup_command_palette()
        self.setup_stats_palette()
        self.setup_map_pane()
        self.setup_df_pane()
        self.setup_statusbar()
        self.retranslateUi(OuterWindow)

        OuterWindow.setCentralWidget(self.centralwidget)
        QtCore.QMetaObject.connectSlotsByName(OuterWindow)

    def setup_df_pane(self):
        """
            Set up the pane that displays the dataframe.

            Returns
            -------
                None
        """
        # Create the layout manager and set the size of the pane.
        self.dflayoutmanager = QtWidgets.QWidget(self.centralwidget)
        self.dflayoutmanager.setObjectName("verticalLayoutWidget")
        self.DFPane = QtWidgets.QVBoxLayout(self.dflayoutmanager)
        self.DFPane.setContentsMargins(0, 0, 0, 0)
        self.DFPane.setObjectName("DFPane")
        self.vlayout.addWidget(self.dflayoutmanager, 3, 1, 1, 2)

    def setup_stats_palette(self):
        """
            Set up the pane that displays the statistics.

            Returns
            -------
                None
        """
        # Create the layout manager and set the size of the pane.
        self.statslayoutmanager = QtWidgets.QWidget(self.centralwidget)
        self.statslayoutmanager.setObjectName("gridLayoutWidget")
        self.StatsPane = QtWidgets.QVBoxLayout(self.statslayoutmanager)
        self.StatsPane.setContentsMargins(0, 0, 0, 0)
        self.StatsPane.setObjectName("StatsPane")

        self.vlayout.addWidget(self.statslayoutmanager, 0, 3, 4, 1)

    def setup_map_pane(self):
        """
            Set up the pane that displays the map.

            Returns
            -------
                None
        """
        # Add the layout manager and then finally set the size and shape.
        self.maplayoutmanager = QtWidgets.QWidget(self.centralwidget)
        self.maplayoutmanager.setObjectName("gridLayoutWidget_2")
        self.MapPane = QtWidgets.QGridLayout(self.maplayoutmanager)
        self.MapPane.setContentsMargins(0, 0, 0, 0)
        self.MapPane.setObjectName("MapPane")

        self.open_btn = QtWidgets.QPushButton("Open File")
        self.open_btn.clicked.connect(self.open_file)
        self.open_btn.resize(150, 50)
        self.MapPane.addWidget(self.open_btn)

        self.vlayout.addWidget(self.maplayoutmanager, 0, 1, 3, 2)

    def setup_command_palette(self):
        """
            Set up the pane that displays the command palette.

            Returns
            -------
                None
        """
        # Create the layout manager and set the size of the pane.
        self.cmdlayoutmanager = QtWidgets.QWidget(self.centralwidget)
        self.cmdlayoutmanager.setObjectName("verticalLayoutWidget_2")
        self.CommandPalette = QtWidgets.QVBoxLayout(self.cmdlayoutmanager)
        self.CommandPalette.setContentsMargins(0, 0, 0, 0)
        self.CommandPalette.setObjectName("CommandPalette")

        # ---------------------------------------------------------------------------------- #
        # Declare the list containing the features and then individual lists of all features.
        feature_types = [
            'Kinematic Features', 'Temporal Features', 'Filtering', 'Interpolation', 'Statistics'
        ]

        newVLayout = QtWidgets.QVBoxLayout()
        # ------------------- Feature Selection List ---------------------- #
        self.featureType = QtWidgets.QComboBox()
        self.featureType.addItems(sorted(feature_types))
        self.featureType.setFont(QtGui.QFont('Tahoma', 12))
        self.featureType.currentIndexChanged.connect(self.add_tree_options)
        newVLayout.addWidget(self.featureType)

        # ------------------- Multi Selection Widget ----------------------#
        self.featureListWidget = QtWidgets.QListWidget()
        self.featureListWidget.setSizePolicy(QSizePolicy(QSizePolicy.Ignored, QSizePolicy.Ignored))
        filt_features = [
            'Hampel Filter', 'Remove Duplicates', 'By Trajectory ID', 'By Bounding Box', 'By Date',
            'By DateTime', 'By Maximum Speed', 'By Minimum Speed', 'By Minimum Consecutive Distance',
            'By Maximum Consecutive Distance', 'By Maximum Distance and Speed', 'By Minimum Distance and Speed',
            'Remove Outliers by Consecutive Distance', 'Remove Outliers by Consecutive Speed',
            'Remove Trajectories with Less Points',
        ]

        self.featureListWidget.addItems(filt_features)
        self.featureListWidget.setFont(QtGui.QFont('Tahoma', 12))
        self.featureListWidget.setUniformItemSizes(True)
        self.featureListWidget.item(0).setSelected(True)
        newVLayout.addWidget(self.featureListWidget)

        # ------------- Add the run commands button. --------------------- #
        self.runStatsBtn = QtWidgets.QPushButton("Run")
        self.runStatsBtn.toggle()
        self.runStatsBtn.setFont(QtGui.QFont('Tahoma', 12))
        self.runStatsBtn.setEnabled(False)
        self.runStatsBtn.clicked.connect(lambda: self.handler.run_command())
        newVLayout.addWidget(self.runStatsBtn)
        self.CommandPalette.addLayout(newVLayout)

        self.vlayout.addWidget(self.cmdlayoutmanager, 0, 0, 4, 1)

    def add_tree_options(self):
        ip_features = [
            'Linear Interpolation', 'Cubic Interpolation', 'Kinematic Interpolation',
            'Random-Walk Interpolation'
        ]

        kinematic_features = [
            'All Kinematic Features', 'Distance', 'Distance from Start', 'Point within Range',
            'Distance from Co-ordinates', 'Speed', 'Acceleration', 'Jerk', 'Bearing', 'Bearing Rate',
            'Rate of Bearing Rate',
        ]

        filt_features = [
            'Hampel Filter', 'Remove Duplicates', 'By Trajectory ID', 'By Bounding Box', 'By Date',
            'By DateTime', 'By Maximum Speed', 'By Minimum Speed', 'By Minimum Consecutive Distance',
            'By Maximum Consecutive Distance', 'By Maximum Distance and Speed', 'By Minimum Distance and Speed',
            'Remove Outliers by Consecutive Distance', 'Remove Outliers by Consecutive Speed',
            'Remove Trajectories with Less Points',
        ]

        stat_features = [
            'Segment Trajectories', 'Generate Kinematic Statistics',
            'Pivot Statistics DF'
        ]

        temporal_features = [
            'All Temporal Features', 'Date', 'Time',
            'Day of the Week', 'Weekend Indicator', 'Time of Day',
        ]
        self.featureListWidget.clear()

        # Change the options of the list as per the current
        # feature selection.
        if self.featureType.currentIndex() == 0:
            self.featureListWidget.addItems(filt_features)
        elif self.featureType.currentIndex() == 1:
            self.featureListWidget.addItems(ip_features)
        elif self.featureType.currentIndex() == 2:
            self.featureListWidget.addItems(kinematic_features)
        elif self.featureType.currentIndex() == 3:
            self.featureListWidget.addItems(stat_features)
        else:
            self.featureListWidget.addItems(temporal_features)

        self.featureListWidget.setFont(QtGui.QFont('Tahoma', 12))
        self.featureListWidget.item(0).setSelected(True)
        self.featureListWidget.setUniformItemSizes(True)

    def setup_menubar(self):
        """
            Create the menu bar of the window.

            Returns
            -------
                None
        """
        # Create the Menu Bar.
        self.MenuBar = QtWidgets.QMenuBar(self.OuterWindow)
        self.MenuBar.setObjectName("MenuBar")

        # Create the File Menu.
        self.FileMenu = QtWidgets.QMenu(self.MenuBar)
        self.FileMenu.setObjectName("FileMenu")

        # Create the open button of the File Menu.
        self.OpenButton = QtWidgets.QAction(self.OuterWindow)
        self.OpenButton.triggered.connect(self.open_file)
        self.OpenButton.setObjectName("OpenButton")
        self.OpenButton.setShortcut("Ctrl+O" if sys.platform != 'darwin' else "cmd+O")
        self.FileMenu.addAction(self.OpenButton)

        # Create the save button of the File Menu.
        self.SaveButton = QtWidgets.QAction(self.OuterWindow)
        self.SaveButton.triggered.connect(self.save_file)
        self.SaveButton.setObjectName("SaveButton")
        self.SaveButton.setShortcut("Ctrl+S" if sys.platform != 'darwin' else "cmd+S")
        self.FileMenu.addAction(self.SaveButton)

        self.FileMenu.addSeparator()
        # Create the quit button of the File Menu and add action listener.
        self.QuitButton = QtWidgets.QAction(self.OuterWindow)
        self.QuitButton.triggered.connect(QtWidgets.qApp.quit)
        self.QuitButton.setShortcut("Alt+F4" if sys.platform != 'darwin' else 'cmd+W')
        self.QuitButton.setObjectName("QuitButton")
        self.FileMenu.addAction(self.QuitButton)

        # Create the About Menu.
        self.menuAbout = QtWidgets.QMenu(self.MenuBar)
        self.menuAbout.setObjectName("menuAbout")

        # Add the Version Info Button to the About Menu and add the action listener.
        self.VersionInfoButton = QtWidgets.QAction(self.OuterWindow)
        self.VersionInfoButton.triggered.connect(self.version_button_clicked)
        self.VersionInfoButton.setShortcut("F1")
        self.VersionInfoButton.setObjectName("VersionInfoButton")
        self.menuAbout.addAction(self.VersionInfoButton)

        self.MenuBar.addAction(self.FileMenu.menuAction())
        self.MenuBar.addAction(self.menuAbout.menuAction())
        self.OuterWindow.setMenuBar(self.MenuBar)

    def retranslateUi(self, OuterWindow):
        """ Auto Generated method by PyQt Designer."""
        _translate = QtCore.QCoreApplication.translate
        OuterWindow.setWindowTitle(_translate("OuterWindow", "PTRAIL - A Trajectory Preprocessing Tool"))
        self.FileMenu.setTitle(_translate("OuterWindow", "File"))
        self.menuAbout.setTitle(_translate("OuterWindow", "About"))
        self.OpenButton.setText(_translate("OuterWindow", "Open"))
        self.SaveButton.setText(_translate("OuterWindow", "Export"))
        self.QuitButton.setText(_translate("OuterWindow", "Quit"))
        self.VersionInfoButton.setText(_translate("OuterWindow", "Version Info"))

    def add_df_controller(self):
        # Create a smaller box layout.
        self.dfController = QtWidgets.QHBoxLayout()

        # # Create a trajectory View selector.
        # self.dfView = QtWidgets.QComboBox()
        # self.dfView.addItems(['Point-Based View', 'Segment-Based View'])
        # self.dfController.addWidget(self.dfView)

        # Create an export button for the dataset.
        self.exportBtn = QtWidgets.QPushButton("Export Dataframe")
        self.exportBtn.clicked.connect(self.save_file)
        self.dfController.addWidget(self.exportBtn)

        # Add this mini panel to the main DF panel.
        self.DFPane.addLayout(self.dfController)

    def setup_statusbar(self):
        self.statusBar = QtWidgets.QStatusBar()
        self.statusBar.setStyleSheet("border :1px solid grey;")
        self.statusBar.setFont(QtGui.QFont('Tahoma', 12))
        self.OuterWindow.setStatusBar(self.statusBar)

    def open_file(self):
        """
            Open the file and load the dataframe to perform operations.

            Returns
            -------
                None
        """
        # Create the file selection dialog box.
        options = QtWidgets.QFileDialog.Options()

        # Limit the user to select only CSV files to load.
        fileName, _ = QtWidgets.QFileDialog.getOpenFileName(None, "Load Dataset", "",
                                                            ";csv Files (*.csv)", options=options)

        # If the user selects a file, load it into a pandas dataframe and print its statistics.
        if fileName is not None and fileName != '':
            self.handler = GuiHandler(fileName, self)
        else:
            pass

    def save_file(self):
        """
            Save the dataframe to a .csv file.

            Returns
            -------
                None
        """
        # Create the file selection dialog box.
        options = QtWidgets.QFileDialog.Options()

        # Limit the user to select only CSV files to load.
        file, check = QtWidgets.QFileDialog.getSaveFileName(None, "Save File", "",
                                                            "Csv Files (*.csv)", options=options)

        # If the user has decided to save the file, save the dataframe to the
        # specified location and display the success message.
        if file and self.handler._data is not None:
            try:
                self.handler._data.to_csv(file + '.csv')
                msg = QtWidgets.QMessageBox()
                msg.setIcon(QtWidgets.QMessageBox.Information)
                msg.setWindowTitle("Success")
                msg.setText("File Saved Successfully!")
                msg.exec()
            except PermissionError:
                msg = QtWidgets.QMessageBox()
                msg.setIcon(QtWidgets.QMessageBox.Critical)
                msg.setWindowTitle("Error")
                msg.setText("File could not be saved. Please try again!")
                msg.exec()
        else:
            pass

    def version_button_clicked(self):
        """
            Show the version info of the application.

            Returns
            -------
                None
        """
        msg = QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Information)
        msg.setWindowTitle("Version Info")
        msg.setText("Version: 0.6.3 Beta \n"
                    "Authors: Yaksh J Haranwala, Salman Haidri")
        msg.exec()



"""
    This File contains TrajectoryPlotter for the Starkey dataset. An interactive experience
    is added to this plot in order to view the trajectory of an individual or multiple animals
    together.

    Warning
    -------
        The visualizations in this module are currently developed with a focus around the
        starkey.csv data as it has been developed as a side project by the developers. It
        will further be integrated into the library as a general class of visualizers in
        the time to come.

    | Authors: Yaksh J Haranwala
"""
import random
import folium

import ipywidgets as widgets
from IPython.core.display import display

from ptrail.core.TrajectoryDF import PTRAILDataFrame
import ptrail.utilities.constants as const


class TrajectoryPlotter:
    # Class variables to handle the ipywidgets.
    _dataset = None
    _weight = None
    _opacity = None
    _selector = None
    _animal = None

    @staticmethod
    def _create_multi_select(dataset, animal):
        """
            Create the multiple selection widget.
            Parameters
            ----------
                dataset: PTRAILDataFrame
                    The dataset from which the IDs are to be selected.
                animal: str
                    The animal for which the list is to be presented.

            Returns
            -------
                ipywidgets.widgets.SelectMultiple
                    Multiple selection widget.

        """
        dataset = dataset.reset_index()

        # Select the animal based on the parameter passed.
        to_select = None
        if animal.lower() == 'deer':
            to_select = dataset.loc[dataset.Species == 0, 'traj_id'].unique()
        elif animal.lower() == 'elk':
            to_select = dataset.loc[dataset.Species == 1, 'traj_id'].unique()
        elif animal.lower() == 'cattle':
            to_select = dataset.loc[dataset.Species == 2, 'traj_id'].unique()

        # Create the multi select widget and return it.
        ids_ = widgets.SelectMultiple(options=to_select, value=(to_select[0], ),
                                      description="Trajectory ID: ", disabled=False)
        return ids_

    @staticmethod
    def _create_radio(value="Cattle"):
        """
            Create the radio button for selecting the animal.

            Returns
            -------
                ipywidgets.widget.RadioButtons
                    The Radio button for selecting the animal.
        """
        radio = widgets.RadioButtons(options=['Cattle', 'Deer', 'Elk'],
                                     value=value, description='Animal: ',
                                     disabled=False)
        return radio

    @staticmethod
    def _filter_dataset(dataset, _id):
        """
            Filter the dataset based on the ids given by the method below.

            Parameters
            ----------
                dataset: PTRAILDataFrame
                    The dataset from which the data is to be filtered.
                _id: tuple
                    The tuple containing the IDs that are required.

            Return
            ------
                PTRAILDataFrame
                    The filtered dataframe.
        """
        filtered_df = dataset.reset_index().loc[dataset.reset_index()['traj_id'].isin(_id)]
        return PTRAILDataFrame(filtered_df.reset_index(), const.LAT, const.LONG, const.DateTime, const.TRAJECTORY_ID)

    @staticmethod
    def _plot(value):
        """
            Show the folium map and plot the trajectories on it.

            Parameters
            ----------
                value: ipywidgets.widget.MultiSelect
                    The Trajectory selector.

            Returns
            -------
                None

        """
        # Register the observer for the animal radio buttons.
        TrajectoryPlotter._animal.observe(TrajectoryPlotter._animal_observe, names="value")

        # Filter the dataset according the values of the widgets above.
        dataset = TrajectoryPlotter._filter_dataset(TrajectoryPlotter._dataset, value)

        # The southwest and northeast bounds.
        sw = dataset[['lat', 'lon']].min().values.tolist()
        ne = dataset[['lat', 'lon']].max().values.tolist()

        # Create a map with the initial point.
        map_ = folium.Map(location=(dataset.latitude[0], dataset.longitude[0]),
                          disable_3d=True, zoom_start=True)

        ids_ = list(dataset.traj_id.value_counts().keys())
        colors = ["#" + ''.join([random.choice('123456789BCDEF') for j in range(6)])
                  for i in range(len(ids_))]

        for i in range(len(ids_)):
            # First, filter out the smaller dataframe.
            small_df = dataset.reset_index().loc[dataset.reset_index()[const.TRAJECTORY_ID] == ids_[i],
                                                 [const.LAT, const.LONG]]

            # Then, create (lat, lon) pairs for the data points.
            locations = []
            for j in range(len(small_df)):
                locations.append((small_df['lat'].iloc[j], small_df['lon'].iloc[j]))

            # Create text frame.
            iframe = folium.IFrame(f'<font size="1px">Trajectory ID: {ids_[i]} ' + '<br>' +
                                   f'Latitude: {locations[0][0]}' + '<br>' +
                                   f'Longitude: {locations[0][1]} </font>')

            # Create start and end markers for the trajectory.
            popup = folium.Popup(iframe, min_width=180, max_width=200, max_height=75)

            folium.Marker([small_df['lat'].iloc[0], small_df['lon'].iloc[0]],
                          color='green',
                          popup=popup,
                          marker_color='green',
                          icon=folium.Icon(icon_color='green', icon='circle', prefix='fa')).add_to(map_)

            # Create text frame.
            iframe = folium.IFrame(f'<font size="1px">Trajectory ID: {ids_[i]} ' + '<br>' +
                                   f'Latitude: {locations[0][0]}' + '<br>' +
                                   f'Longitude: {locations[0][1]} </font>')

            # Create start and end markers for the trajectory.
            popup = folium.Popup(iframe, min_width=180, max_width=200, max_height=75)

            folium.Marker([small_df['lat'].iloc[-1], small_df['lon'].iloc[-1]],
                          color='green',
                          popup=popup,
                          marker_color='red',
                          icon=folium.Icon(icon_color='red', icon='circle', prefix='fa')).add_to(map_)

            # Add trajectory to map.
            folium.PolyLine(locations,
                            color=colors[i],
                            weight=TrajectoryPlotter._weight,
                            opacity=TrajectoryPlotter._opacity).add_to(map_)

        # Fit the map within its bounds and return it.
        map_.fit_bounds([sw, ne])
        display(map_)

    @staticmethod
    def _animal_observe(change):
        """
            This is the observer that changes the multi selection list when the
            value of the Animal radio button is changed.

            Parameters
            ----------
                change: dict
                    The dictionary that contains the new and old values of the
                    radio button.

            Returns
            -------
                None
        """
        to_select = None
        if change['new'].lower() == 'deer':
            to_select = TrajectoryPlotter._dataset.reset_index().loc[
                TrajectoryPlotter._dataset.reset_index().Species == 0, 'traj_id'].unique()
        elif change['new'].lower() == 'elk':
            to_select = TrajectoryPlotter._dataset.reset_index().loc[
                TrajectoryPlotter._dataset.reset_index().Species == 1, 'traj_id'].unique()
        elif change['new'].lower() == 'cattle':
            to_select = TrajectoryPlotter._dataset.reset_index().loc[
                TrajectoryPlotter._dataset.reset_index().Species == 2, 'traj_id'].unique()

        # Based on the new selection, modify the options of the MultiSelector.
        TrajectoryPlotter._selector.options = to_select

        # Also, modify the value of the MultiSelector.
        # WARNING: Do not remove the comma since teh value expects a tuple!
        TrajectoryPlotter._selector.value = to_select[0],

    @staticmethod
    def show_trajectories(dataset, weight: float = 3, opacity: float = 0.8):
        """
            Use folium to plot the trajectory on a map.

            Parameters
            ----------
                dataset:

                weight: float
                    The weight of the trajectory line on the map.
                opacity: float
                    The opacity of the trajectory line on the map.

            Returns
            -------
                folium.folium.Map
                    The map with plotted trajectory.
        """
        # Set the dataset, weight and opacity as the class variables.
        TrajectoryPlotter._dataset = dataset
        TrajectoryPlotter._weight = weight
        TrajectoryPlotter._opacity = opacity

        # Create the radio button.
        TrajectoryPlotter._animal = TrajectoryPlotter._create_radio()

        # Create the multi selection button.
        TrajectoryPlotter._selector = TrajectoryPlotter._create_multi_select(TrajectoryPlotter._dataset,
                                                                             TrajectoryPlotter._animal.value)

        # Create the widgets.
        ie = widgets.interactive_output(TrajectoryPlotter._plot, {'value': TrajectoryPlotter._selector})

        # Display the multi selector and the radio buttons next to each other.
        display(widgets.HBox([TrajectoryPlotter._animal, TrajectoryPlotter._selector]), ie)
